from typing import Dict, Optional

import numpy as np
import torch
import torch_geometric
from torch import nn

from .molecule import batch_to_dict
from .transformer import Transformer, TransformerLayer
from .utilities import NeverRun, get_activation, get_rotations


class CentralSplitter(torch.nn.Module):
    def __init__(self):
        super(CentralSplitter, self).__init__()

    def forward(self, features, central_species):
        all_species = np.unique(central_species)
        all_species = [str(specie) for specie in all_species]

        result = {}
        for specie in all_species:
            result[specie] = {}

        for key, value in features.items():
            for specie in all_species:
                mask_now = central_species == int(specie)
                result[specie][key] = value[mask_now]
        return result


class CentralUniter(torch.nn.Module):
    def __init__(self):
        super(CentralUniter, self).__init__()

    def forward(self, features, central_species):
        all_species = np.unique(central_species)
        all_species = [str(specie) for specie in all_species]
        specie = all_species[0]

        shapes = {}
        for key, value in features[specie].items():
            now = list(value.shape)
            now[0] = 0
            shapes[key] = now

        device = None
        for specie in all_species:
            for key, _value in features[specie].items():
                num = features[specie][key].shape[0]
                device = features[specie][key].device
                shapes[key][0] += num

        result = {
            key: torch.empty(shape, dtype=torch.get_default_dtype()).to(device)
            for key, shape in shapes.items()
        }

        for specie in features.keys():
            for key, _value in features[specie].items():
                mask = int(specie) == central_species
                result[key][mask] = features[specie][key]

        return result


def cutoff_func(grid: torch.Tensor, r_cut: float, delta: float):
    mask_bigger = grid >= r_cut
    mask_smaller = grid <= r_cut - delta
    grid = (grid - r_cut + delta) / delta
    f = 1 / 2.0 + torch.cos(np.pi * grid) / 2.0

    f[mask_bigger] = 0.0
    f[mask_smaller] = 1.0
    return f


class CartesianTransformer(torch.nn.Module):
    def __init__(
        self,
        hypers,
        d_model,
        n_head,
        dim_feedforward,
        n_layers,
        dropout,
        n_atomic_species,
        add_central_token,
        is_first,
    ):
        super(CartesianTransformer, self).__init__()
        self.hypers = hypers
        self.is_first = is_first
        self.trans_layer = TransformerLayer(
            d_model=d_model,
            n_heads=n_head,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            activation=get_activation(hypers),
            transformer_type=hypers.TRANSFORMER_TYPE,
        )
        self.trans = Transformer(self.trans_layer, num_layers=n_layers)

        if hypers.USE_ONLY_LENGTH:
            input_dim = 1
        else:
            input_dim = 3
            if hypers.USE_LENGTH:
                input_dim += 1

        if hypers.USE_ADDITIONAL_SCALAR_ATTRIBUTES:
            input_dim += hypers.SCALAR_ATTRIBUTES_SIZE

        if hypers.R_EMBEDDING_ACTIVATION:
            self.r_embedding = nn.Sequential(
                nn.Linear(input_dim, d_model), get_activation(hypers)
            )
        else:
            self.r_embedding = nn.Linear(input_dim, d_model)

        if hypers.BLEND_NEIGHBOR_SPECIES and (not is_first):
            n_merge = 3
        else:
            n_merge = 2

        self.compress = None
        if hypers.COMPRESS_MODE == "linear":
            self.compress = nn.Linear(n_merge * d_model, d_model)
        if hypers.COMPRESS_MODE == "mlp":
            self.compress = nn.Sequential(
                nn.Linear(n_merge * d_model, d_model),
                get_activation(hypers),
                nn.Linear(d_model, d_model),
            )
        if self.compress is None:
            raise ValueError("unknown compress mode")

        self.neighbor_embedder = NeverRun()  # for torchscript
        if hypers.BLEND_NEIGHBOR_SPECIES and (not is_first):
            self.neighbor_embedder = nn.Embedding(n_atomic_species + 1, d_model)

        self.add_central_token = add_central_token

        self.central_embedder = NeverRun()  # for torchscript
        self.central_scalar_embedding = NeverRun()  # for torchscript
        self.central_compress = NeverRun()  # for torchscript

        if add_central_token:
            self.central_embedder = nn.Embedding(n_atomic_species + 1, d_model)
            if hypers.USE_ADDITIONAL_SCALAR_ATTRIBUTES:
                if hypers.R_EMBEDDING_ACTIVATION:
                    self.central_scalar_embedding = nn.Sequential(
                        nn.Linear(hypers.SCALAR_ATTRIBUTES_SIZE, d_model),
                        get_activation(hypers),
                    )
                else:
                    self.central_scalar_embedding = nn.Linear(
                        hypers.SCALAR_ATTRIBUTES_SIZE, d_model
                    )

                if hypers.COMPRESS_MODE == "linear":
                    self.central_compress = nn.Linear(2 * d_model, d_model)
                if hypers.COMPRESS_MODE == "mlp":
                    self.central_compress = nn.Sequential(
                        nn.Linear(2 * d_model, d_model),
                        get_activation(hypers),
                        nn.Linear(d_model, d_model),
                    )

        # assign hypers one by one for torch.script
        self.USE_LENGTH = hypers.USE_LENGTH
        self.BLEND_NEIGHBOR_SPECIES = hypers.BLEND_NEIGHBOR_SPECIES
        self.USE_ADDITIONAL_SCALAR_ATTRIBUTES = hypers.USE_ADDITIONAL_SCALAR_ATTRIBUTES
        self.USE_ONLY_LENGTH = hypers.USE_ONLY_LENGTH
        self.R_CUT = hypers.R_CUT
        self.CUTOFF_DELTA = hypers.CUTOFF_DELTA

    def forward(self, batch_dict: Dict[str, torch.Tensor]):
        x = batch_dict["x"]

        if self.USE_LENGTH:
            neighbor_lengths = torch.sqrt(torch.sum(x**2, dim=2) + 1e-15)[:, :, None]
        else:
            neighbor_lengths = torch.empty(
                0, device=x.device, dtype=x.dtype
            )  # for torch script

        central_species = batch_dict["central_species"]
        neighbor_species = batch_dict["neighbor_species"]
        input_messages = batch_dict["input_messages"]
        mask = batch_dict["mask"]
        nums = batch_dict["nums"]

        if self.BLEND_NEIGHBOR_SPECIES and (not self.is_first):
            neighbor_embedding = self.neighbor_embedder(neighbor_species)
        else:
            neighbor_embedding = torch.empty(
                0, device=x.device, dtype=x.dtype
            )  # for torch script

        if self.USE_ADDITIONAL_SCALAR_ATTRIBUTES:
            neighbor_scalar_attributes = batch_dict["neighbor_scalar_attributes"]
            central_scalar_attributes = batch_dict["central_scalar_attributes"]
        else:
            neighbor_scalar_attributes = torch.empty(
                0, device=x.device, dtype=x.dtype
            )  # for torch script
            central_scalar_attributes = torch.empty(
                0, device=x.device, dtype=x.dtype
            )  # for torch script

        initial_n_tokens = x.shape[1]
        max_number = int(torch.max(nums))

        if self.USE_ONLY_LENGTH:
            coordinates = [neighbor_lengths]
        else:
            coordinates = [x]
            if self.USE_LENGTH:
                coordinates.append(neighbor_lengths)

        if self.USE_ADDITIONAL_SCALAR_ATTRIBUTES:
            coordinates.append(neighbor_scalar_attributes)
        coordinates = torch.cat(coordinates, dim=2)
        coordinates = self.r_embedding(coordinates)

        if self.BLEND_NEIGHBOR_SPECIES and (not self.is_first):
            tokens = torch.cat([coordinates, neighbor_embedding, input_messages], dim=2)
        else:
            tokens = torch.cat([coordinates, input_messages], dim=2)

        tokens = self.compress(tokens)

        if self.add_central_token:
            central_specie_embedding = self.central_embedder(central_species)
            if self.USE_ADDITIONAL_SCALAR_ATTRIBUTES:
                central_scalar_embedding = self.central_scalar_embedding(
                    central_scalar_attributes
                )
                central_token = torch.cat(
                    [central_specie_embedding, central_scalar_embedding], dim=1
                )
                central_token = self.central_compress(central_token)
            else:
                central_token = central_specie_embedding

            tokens = torch.cat([central_token[:, None, :], tokens], dim=1)

            submask = torch.zeros(mask.shape[0], dtype=torch.bool).to(mask.device)
            total_mask = torch.cat([submask[:, None], mask], dim=1)

            lengths = torch.sqrt(torch.sum(x * x, dim=2) + 1e-16)
            multipliers = cutoff_func(lengths, self.R_CUT, self.CUTOFF_DELTA)
            sub_multipliers = torch.ones(mask.shape[0], device=mask.device)
            multipliers = torch.cat([sub_multipliers[:, None], multipliers], dim=1)
            multipliers[total_mask] = 0.0

            multipliers = multipliers[:, None, :]
            multipliers = multipliers.repeat(1, multipliers.shape[2], 1)

            output_messages = self.trans(
                tokens[:, : (max_number + 1), :],
                multipliers=multipliers[:, : (max_number + 1), : (max_number + 1)],
            )
            if max_number < initial_n_tokens:
                padding = torch.zeros(
                    output_messages.shape[0],
                    initial_n_tokens - max_number,
                    output_messages.shape[2],
                    device=output_messages.device,
                )
                output_messages = torch.cat([output_messages, padding], dim=1)

            return {
                "output_messages": output_messages[:, 1:, :],
                "central_token": output_messages[:, 0, :],
            }
        else:
            lengths = torch.sqrt(torch.sum(x * x, dim=2) + 1e-16)

            multipliers = cutoff_func(lengths, self.R_CUT, self.CUTOFF_DELTA)
            multipliers[mask] = 0.0

            multipliers = multipliers[:, None, :]
            multipliers = multipliers.repeat(1, multipliers.shape[2], 1)

            output_messages = self.trans(
                tokens[:, :max_number, :],
                multipliers=multipliers[:, :max_number, :max_number],
            )
            if max_number < initial_n_tokens:
                padding = torch.zeros(
                    output_messages.shape[0],
                    initial_n_tokens - max_number,
                    output_messages.shape[2],
                    device=output_messages.device,
                )
                output_messages = torch.cat([output_messages, padding], dim=1)

            return {"output_messages": output_messages}


class CentralSpecificModel(torch.nn.Module):
    def __init__(self, models):
        super(CentralSpecificModel, self).__init__()
        self.models = torch.nn.ModuleDict(models)
        self.splitter = CentralSplitter()
        self.uniter = CentralUniter()

    def forward(self, batch_dict):
        central_indices = batch_dict["central_species"].data.cpu().numpy()
        splitted = self.splitter(batch_dict, central_indices)

        result = {}
        for key in splitted.keys():
            result[str(key)] = self.models[str(key)](splitted[key])

        result = self.uniter(result, central_indices)
        return result


class LinearWithFeatures(torch.nn.Module):
    def __init__(self, *args, **kwargs):
        super(LinearWithFeatures, self).__init__()
        self.linear = nn.Linear(*args, **kwargs)

    def forward(self, x):
        return {
            "features": x,
            "output": self.linear(x),
        }


class Head(torch.nn.Module):
    def __init__(self, hypers, n_in, n_neurons):
        super(Head, self).__init__()
        self.hypers = hypers
        self.nn = nn.Sequential(
            nn.Linear(n_in, n_neurons),
            get_activation(hypers),
            nn.Linear(n_neurons, n_neurons),
            get_activation(hypers),
            LinearWithFeatures(n_neurons, hypers.D_OUTPUT),
        )

    def forward(self, batch_dict: Dict[str, torch.Tensor]):
        pooled = batch_dict["pooled"]
        outputs = self.nn(pooled)
        return {
            "atomic_predictions": outputs["output"],
            "features": outputs["features"],
        }


class CentralTokensPredictor(torch.nn.Module):
    def __init__(self, hypers, head):
        super(CentralTokensPredictor, self).__init__()
        self.head = head
        self.hypers = hypers

    def forward(self, central_tokens: torch.Tensor, central_species: torch.Tensor):
        head_output = self.head(
            {"pooled": central_tokens, "central_species": central_species}
        )
        predictions = head_output["atomic_predictions"]
        features = head_output["features"]
        return {
            "atomic_predictions": predictions,
            "features": features,
        }


class MessagesPredictor(torch.nn.Module):
    def __init__(self, hypers, head):
        super(MessagesPredictor, self).__init__()
        self.head = head
        self.AVERAGE_POOLING = hypers.AVERAGE_POOLING

    def forward(
        self,
        messages: torch.Tensor,
        mask: torch.Tensor,
        nums: torch.Tensor,
        central_species: torch.Tensor,
        multipliers: torch.Tensor,
    ):
        messages_proceed = messages * multipliers[:, :, None]
        messages_proceed[mask] = 0.0
        if self.AVERAGE_POOLING:
            total_weight = multipliers.sum(dim=1)[:, None]
            pooled = messages_proceed.sum(dim=1) / total_weight
        else:
            pooled = messages_proceed.sum(dim=1)

        predictions = self.head({"pooled": pooled, "central_species": central_species})[
            "atomic_predictions"
        ]
        return predictions


class MessagesBondsPredictor(torch.nn.Module):
    def __init__(self, hypers, head):
        super(MessagesBondsPredictor, self).__init__()
        self.head = head
        self.AVERAGE_BOND_ENERGIES = hypers.AVERAGE_BOND_ENERGIES

    def forward(
        self,
        messages: torch.Tensor,
        mask: torch.Tensor,
        nums: torch.Tensor,
        central_species: torch.Tensor,
        multipliers: torch.Tensor,
    ):
        head_output = self.head(
            {"pooled": messages, "central_species": central_species}
        )

        predictions = head_output["atomic_predictions"]
        features = head_output["features"]

        mask_expanded = mask[..., None].repeat(1, 1, predictions.shape[2])
        predictions = torch.where(mask_expanded, 0.0, predictions)

        mask_expanded_features = mask[..., None].repeat(1, 1, features.shape[2])
        features = torch.where(mask_expanded_features, 0.0, features)

        predictions = predictions * multipliers[:, :, None]
        if self.AVERAGE_BOND_ENERGIES:
            raise NotImplementedError(
                "AVERAGE_BOND_ENERGIES not implemented in the last-layer "
                " feature branch."
            )
            total_weight = multipliers.sum(dim=1)[:, None]
            result = predictions.sum(dim=1) / total_weight
        else:
            result = predictions.sum(dim=1)
            features = features.sum(dim=1)
        return {
            "atomic_predictions": result,
            "features": features,
        }


class PET(torch.nn.Module):
    def __init__(self, hypers, transformer_dropout, n_atomic_species):
        super(PET, self).__init__()
        self.hypers = hypers
        transformer_d_model = hypers.TRANSFORMER_D_MODEL
        transformer_n_head = hypers.TRANSFORMER_N_HEAD
        transformer_dim_feedforward = hypers.TRANSFORMER_DIM_FEEDFORWARD
        transformer_n_layers = hypers.N_TRANS_LAYERS
        n_gnn_layers = hypers.N_GNN_LAYERS
        head_n_neurons = hypers.HEAD_N_NEURONS
        transformers_central_specific = hypers.TRANSFORMERS_CENTRAL_SPECIFIC
        heads_central_specific = hypers.HEADS_CENTRAL_SPECIFIC

        add_central_tokens = []
        for _ in range(hypers.N_GNN_LAYERS - 1):
            add_central_tokens.append(hypers.ADD_TOKEN_FIRST)
        add_central_tokens.append(hypers.ADD_TOKEN_SECOND)

        self.embedding = nn.Embedding(n_atomic_species + 1, transformer_d_model)
        gnn_layers = []
        if transformers_central_specific:
            for layer_index in range(n_gnn_layers):
                if layer_index == 0:
                    is_first = True
                else:
                    is_first = False
                models = {
                    str(i): CartesianTransformer(
                        hypers,
                        transformer_d_model,
                        transformer_n_head,
                        transformer_dim_feedforward,
                        transformer_n_layers,
                        transformer_dropout,
                        n_atomic_species,
                        add_central_tokens[layer_index],
                        is_first,
                    )
                    for i in range(n_atomic_species)
                }

                gnn_layers.append(CentralSpecificModel(models))
        else:
            for layer_index in range(n_gnn_layers):
                if layer_index == 0:
                    is_first = True
                else:
                    is_first = False
                model = CartesianTransformer(
                    hypers,
                    transformer_d_model,
                    transformer_n_head,
                    transformer_dim_feedforward,
                    transformer_n_layers,
                    transformer_dropout,
                    n_atomic_species,
                    add_central_tokens[layer_index],
                    is_first,
                )
                gnn_layers.append(model)

        self.gnn_layers = torch.nn.ModuleList(gnn_layers)

        heads = []
        if heads_central_specific:
            for _ in range(n_gnn_layers):
                models = {
                    str(i): Head(hypers, transformer_d_model, head_n_neurons)
                    for i in range(n_atomic_species)
                }
                heads.append(CentralSpecificModel(models))

            models = {
                str(i): Head(hypers, transformer_d_model, head_n_neurons)
                for i in range(n_atomic_species)
            }
        else:
            for _ in range(n_gnn_layers):
                heads.append(Head(hypers, transformer_d_model, head_n_neurons))

        self.heads = torch.nn.ModuleList(heads)
        self.central_tokens_predictors = torch.nn.ModuleList(
            [CentralTokensPredictor(hypers, head) for head in heads]
        )
        self.messages_predictors = torch.nn.ModuleList(
            [MessagesPredictor(hypers, head) for head in heads]
        )

        if hypers.USE_BOND_ENERGIES:
            bond_heads = []
            if heads_central_specific:
                for _ in range(n_gnn_layers):
                    models = {
                        str(i): Head(hypers, transformer_d_model, head_n_neurons)
                        for i in range(n_atomic_species)
                    }
                    bond_heads.append(CentralSpecificModel(models))

                models = {
                    str(i): Head(hypers, transformer_d_model, head_n_neurons)
                    for i in range(n_atomic_species)
                }
            else:
                for _ in range(n_gnn_layers):
                    bond_heads.append(Head(hypers, transformer_d_model, head_n_neurons))

            self.bond_heads = torch.nn.ModuleList(bond_heads)
            self.messages_bonds_predictors = torch.nn.ModuleList(
                [MessagesBondsPredictor(hypers, head) for head in bond_heads]
            )
        else:
            self.messages_bonds_predictors = torch.nn.ModuleList(
                [NeverRun() for _ in range(n_gnn_layers)]
            )

        self.R_CUT = hypers.R_CUT
        self.CUTOFF_DELTA = hypers.CUTOFF_DELTA
        self.USE_BOND_ENERGIES = hypers.USE_BOND_ENERGIES
        self.TARGET_TYPE = hypers.TARGET_TYPE
        self.TARGET_AGGREGATION = hypers.TARGET_AGGREGATION
        self.N_GNN_LAYERS = hypers.N_GNN_LAYERS
        self.RESIDUAL_FACTOR = hypers.RESIDUAL_FACTOR

    def get_predictions(self, batch_dict: Dict[str, torch.Tensor]):
        x = batch_dict["x"]
        central_species = batch_dict["central_species"]
        neighbor_species = batch_dict["neighbor_species"]
        mask = batch_dict["mask"]
        nums = batch_dict["nums"]

        lengths = torch.sqrt(torch.sum(x * x, dim=2) + 1e-16)
        multipliers = cutoff_func(lengths, self.R_CUT, self.CUTOFF_DELTA)
        multipliers[mask] = 0.0

        neighbors_index = batch_dict["neighbors_index"]
        neighbors_pos = batch_dict["neighbors_pos"]

        batch_dict["input_messages"] = self.embedding(neighbor_species)
        atomic_predictions = torch.zeros(1, dtype=x.dtype, device=x.device)
        last_layer_features = []

        for _layer_index, (
            central_tokens_predictor,
            messages_predictor,
            gnn_layer,
            messages_bonds_predictor,
        ) in enumerate(
            zip(
                self.central_tokens_predictors,
                self.messages_predictors,
                self.gnn_layers,
                self.messages_bonds_predictors,
            )
        ):
            result = gnn_layer(batch_dict)
            output_messages = result["output_messages"]

            new_input_messages = output_messages[neighbors_index, neighbors_pos]
            batch_dict["input_messages"] = self.RESIDUAL_FACTOR * (
                batch_dict["input_messages"] + new_input_messages
            )

            if "central_token" in result.keys():
                predictor_output = central_tokens_predictor(
                    result["central_token"], central_species
                )
                atomic_predictions = (
                    atomic_predictions + predictor_output["atomic_predictions"]
                )
                last_layer_features.append(predictor_output["features"])
            else:
                raise NotImplementedError(
                    "Central token predictor not implemented in the last-layer feature "
                    "branch."
                )
                atomic_predictions = atomic_predictions + messages_predictor(
                    output_messages, mask, nums, central_species, multipliers
                )

            if self.USE_BOND_ENERGIES:
                predictor_output = messages_bonds_predictor(
                    output_messages, mask, nums, central_species, multipliers
                )
                atomic_predictions = (
                    atomic_predictions + predictor_output["atomic_predictions"]
                )
                last_layer_features.append(predictor_output["features"])

        last_layer_features = torch.concatenate(last_layer_features, dim=1)

        if self.TARGET_TYPE == "structural":
            if self.TARGET_AGGREGATION == "sum":
                prediction = torch_geometric.nn.global_add_pool(
                    atomic_predictions, batch=batch_dict["batch"]
                )
                last_layer_features = torch_geometric.nn.global_mean_pool(
                    last_layer_features, batch=batch_dict["batch"]
                )
                return {
                    "prediction": prediction,
                    "last_layer_features": last_layer_features,
                }
            if self.TARGET_AGGREGATION == "mean":
                raise NotImplementedError(
                    "mean aggregation not implemented in the last-layer feature branch."
                )
                return torch_geometric.nn.global_mean_pool(
                    atomic_predictions, batch=batch_dict["batch"]
                )
            raise ValueError("unknown target aggregation")
        if self.TARGET_TYPE == "atomic":
            return {
                "prediction": atomic_predictions,
                "last_layer_features": last_layer_features,
            }
        raise ValueError("unknown target type")

    def forward(
        self,
        batch_dict: Dict[str, torch.Tensor],
        rotations: Optional[torch.Tensor] = None,
    ):
        if rotations is not None:
            x_initial = batch_dict["x"]
            batch_dict["x"] = torch.bmm(x_initial, rotations)
            predictions = self.get_predictions(batch_dict)
            batch_dict["x"] = x_initial
            return predictions
        else:
            return self.get_predictions(batch_dict)


class PETUtilityWrapper(torch.nn.Module):
    """Uncoupling torch.unscriptable logic from the main PET class"""

    def __init__(self, pet_model, global_aug):
        super(PETUtilityWrapper, self).__init__()
        self.pet_model = pet_model
        self.global_aug = global_aug

    def forward(self, batch, augmentation):
        batch_dict = batch_to_dict(batch)
        rotations = None
        if augmentation:
            indices = batch.batch.cpu().data.numpy()
            rotations = torch.tensor(
                get_rotations(indices, global_aug=self.global_aug),
                device=batch.x.device,
                dtype=batch.x.dtype,
            )
        return self.pet_model(batch_dict, rotations)


class PETMLIPWrapper(torch.nn.Module):
    def __init__(self, model, use_energies, use_forces):
        super(PETMLIPWrapper, self).__init__()
        self.model = model
        self.use_energies = use_energies
        self.use_forces = use_forces
        if self.model.pet_model.hypers.D_OUTPUT != 1:
            raise ValueError("D_OUTPUT should be 1 for MLIP; energy is a single scalar")
        if self.model.pet_model.hypers.TARGET_TYPE != "structural":
            raise ValueError("TARGET_TYPE should be structural for MLIP")
        if self.model.pet_model.hypers.TARGET_AGGREGATION != "sum":
            raise ValueError("TARGET_AGGREGATION should be sum for MLIP")

    def get_predictions(self, batch, augmentation):
        predictions = self.model(batch, augmentation=augmentation)
        if predictions["prediction"].shape[-1] != 1:
            raise ValueError("D_OUTPUT should be 1 for MLIP; energy is a single scalar")
        # if predictions.shape[0] != batch.num_graphs:
        #    raise ValueError("model should return a single scalar per structure")
        return {
            "prediction": predictions["prediction"][..., 0],
            "last_layer_features": predictions["last_layer_features"],
        }

    def forward(self, batch, augmentation, create_graph):
        if self.use_forces:
            batch.x.requires_grad = True
            predictions = self.get_predictions(batch, augmentation)["prediction"]
            grads = torch.autograd.grad(
                predictions,
                batch.x,
                grad_outputs=torch.ones_like(predictions),
                create_graph=create_graph,
            )[0]
            neighbors_index = batch.neighbors_index.transpose(0, 1)
            neighbors_pos = batch.neighbors_pos
            grads_messaged = grads[neighbors_index, neighbors_pos]
            grads[batch.mask] = 0.0
            first = grads.sum(dim=1)
            grads_messaged[batch.mask] = 0.0
            second = grads_messaged.sum(dim=1)
        else:
            predictions = self.get_predictions(batch, augmentation)["prediction"]

        result = []
        if self.use_energies:
            result.append(predictions)
        else:
            result.append(None)

        if self.use_forces:
            result.append(first - second)
        else:
            result.append(None)

        return result


class PETLLFWrapper(torch.nn.Module):
    def __init__(self, model, use_energies, use_forces):
        super(PETLLFWrapper, self).__init__()
        self.model = model
        self.use_energies = use_energies
        self.use_forces = use_forces

        self._validate_model_configuration()

    def _validate_model_configuration(self):
        if self.model.pet_model.hypers.D_OUTPUT != 1:
            raise ValueError("D_OUTPUT should be 1 for MLIP; energy is a single scalar")
        if self.model.pet_model.hypers.TARGET_TYPE != "structural":
            raise ValueError("TARGET_TYPE should be structural for MLIP")
        if self.model.pet_model.hypers.TARGET_AGGREGATION != "sum":
            raise ValueError("TARGET_AGGREGATION should be sum for MLIP")

    def get_predictions(self, batch, augmentation):
        predictions = self.model(batch, augmentation=augmentation)
        if predictions["prediction"].shape[-1] != 1:
            raise ValueError("D_OUTPUT should be 1 for MLIP; energy is a single scalar")

        return {
            "prediction": predictions["prediction"][..., 0],
            "last_layer_features": predictions["last_layer_features"],
        }

    def forward(self, batch, augmentation, create_graph):
        if self.use_forces:
            batch.x.requires_grad = True

        run_result = self.get_predictions(batch, augmentation)
        return run_result["last_layer_features"]


class SelfContributionsWrapper(torch.nn.Module):
    def __init__(self, model, self_contributions):
        super(SelfContributionsWrapper, self).__init__()
        self.model = model
        self.register_buffer(
            "self_contributions",
            torch.tensor(self_contributions, dtype=torch.get_default_dtype()),
        )
        if self.model.hypers.TARGET_TYPE == "structural":
            self.TARGET_TYPE = "structural"  # for TorchScript
            if self.model.hypers.TARGET_AGGREGATION == "mean":
                raise ValueError(
                    "self contributions wrapper is made only for sum "
                    "aggregation, not for mean"
                )
        else:
            self.TARGET_TYPE = "atomic"
        if self.model.hypers.D_OUTPUT != 1:
            raise ValueError("self contributions wrapper is made only for D_OUTPUT = 1")

    def forward(self, batch_dict: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        predictions = self.model(batch_dict)
        central_species = batch_dict["central_species"]
        self_contribution_energies = self.self_contributions[central_species][:, None]
        if self.TARGET_TYPE == "structural":
            self_contribution_energies = torch_geometric.nn.global_add_pool(
                self_contribution_energies, batch=batch_dict["batch"]
            )
        output = {
            "prediction": predictions["prediction"] + self_contribution_energies,
            "last_layer_features": predictions["last_layer_features"],
        }
        return output


class FlagsWrapper(torch.nn.Module):
    """For DataParallel"""

    def __init__(self, model):
        super(FlagsWrapper, self).__init__()
        self.model = model
        self.augmentation = None
        self.create_graph = None

    def forward(self, batch):
        return self.model(
            batch, augmentation=self.augmentation, create_graph=self.create_graph
        )



from typing import Dict, Optional

import torch


class LoRALayer(torch.nn.Module):
    def __init__(self, hidden_dim: int, rank: int):
        super(LoRALayer, self).__init__()
        self.hidden_dim = hidden_dim
        self.rank = rank
        self.A = torch.nn.Parameter(torch.randn(hidden_dim, rank))
        self.B = torch.nn.Parameter(torch.randn(rank, hidden_dim))
        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.zeros_(self.A)
        torch.nn.init.xavier_normal_(self.B)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x @ self.A @ self.B


class AttentionBlockWithLoRA(torch.nn.Module):
    def __init__(self, original_module: torch.nn.Module, rank: int, alpha: float):
        super(AttentionBlockWithLoRA, self).__init__()
        self.original_module = original_module
        self.rank = rank
        self.alpha = alpha
        self.hidden_dim = original_module.output_linear.out_features
        self.lora = LoRALayer(self.hidden_dim, self.rank)

    def forward(
        self, x: torch.Tensor, multipliers: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        return self.original_module(x, multipliers) + self.alpha * self.lora(x)


class LoRAWrapper(torch.nn.Module):
    def __init__(self, model: torch.nn.Module, rank: int, alpha: float):
        super(LoRAWrapper, self).__init__()
        self.model = model
        self.hypers = model.hypers
        self.rank = rank
        self.alpha = alpha
        self.hidden_dim = model.hypers.TRANSFORMER_D_MODEL
        self.num_hidden_layers = model.hypers.N_GNN_LAYERS * model.hypers.N_TRANS_LAYERS
        for param in model.parameters():
            param.requires_grad = False
        for gnn_layer in model.gnn_layers:
            for trans_layer in gnn_layer.trans.layers:
                trans_layer.attention = AttentionBlockWithLoRA(
                    trans_layer.attention, self.rank, self.alpha
                )

    def forward(
        self,
        batch_dict: Dict[str, torch.Tensor],
        rotations: Optional[torch.Tensor] = None,
    ):
        return self.model(batch_dict, rotations)



from typing import Dict, Tuple

from omegaconf import DictConfig

from .dataset import Dataset, DiskDataset
from .readers import read_systems, read_targets
from .target_info import TargetInfo


def get_dataset(options: DictConfig) -> Tuple[Dataset, Dict[str, TargetInfo]]:
    """
    Gets a dataset given a configuration dictionary.

    The system and targets in the dataset are read from one or more
    files, as specified in ``options``.

    :param options: the configuration options for the dataset.
        This configuration dictionary must contain keys for both the
        systems and targets in the dataset.

    :returns: A tuple containing a ``Dataset`` object and a
        ``Dict[str, TargetInfo]`` containing additional information (units,
        physical quantities, ...) on the targets in the dataset
    """

    if options["systems"]["read_from"].endswith(".zip"):  # disk dataset
        dataset = DiskDataset(options["systems"]["read_from"])
        target_info_dictionary = dataset.get_target_info(options["targets"])
    else:
        systems = read_systems(
            filename=options["systems"]["read_from"],
            reader=options["systems"]["reader"],
        )
        targets, target_info_dictionary = read_targets(conf=options["targets"])
        dataset = Dataset.from_dict({"system": systems, **targets})

    return dataset, target_info_dictionary



import torch
from metatensor.torch.atomistic import ModelOutput, System

from metatrain.experimental.nativepet import NativePET
from metatrain.utils.data import DatasetInfo
from metatrain.utils.data.target_info import (
    get_energy_target_info,
)
from metatrain.utils.neighbor_lists import get_system_with_neighbor_lists

from . import MODEL_HYPERS


def test_autograd_positions():
    """Tests the basic functionality of the forward pass of the model."""
    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={
            "energy": get_energy_target_info({"quantity": "energy", "unit": "eV"})
        },
    )

    model = NativePET(MODEL_HYPERS, dataset_info)
    model = model.to(torch.float64)

    positions = torch.tensor(
        [[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]], dtype=torch.float64, requires_grad=True
    )

    def compute(positions):
        system = System(
            types=torch.tensor([6, 6]),
            positions=positions,
            cell=torch.eye(3, dtype=torch.float64),
            pbc=torch.tensor([True, True, True]),
        )

        system = get_system_with_neighbor_lists(
            system, model.requested_neighbor_lists()
        )

        outputs = {"energy": ModelOutput(per_atom=False)}
        output = model([system], outputs)
        energy = output["energy"].block().values.sum()
        return energy

    assert torch.autograd.gradcheck(compute, positions, fast_mode=True)
    assert torch.autograd.gradgradcheck(compute, positions, fast_mode=True)

    if torch.cuda.is_available():
        positions_cuda = positions.detach().cuda().requires_grad_(True)
        assert torch.autograd.gradcheck(compute, positions_cuda, fast_mode=True)
        assert torch.autograd.gradgradcheck(compute, positions_cuda, fast_mode=True)


def test_autograd_cell():
    """Tests the basic functionality of the forward pass of the model."""
    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={
            "energy": get_energy_target_info({"quantity": "energy", "unit": "eV"})
        },
    )

    model = NativePET(MODEL_HYPERS, dataset_info)
    model = model.to(torch.float64)

    cell = torch.eye(3, dtype=torch.float64, requires_grad=True)

    def compute(cell):
        system = System(
            types=torch.tensor([6, 6]),
            positions=torch.tensor(
                [[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]],
                dtype=torch.float64,
                requires_grad=True,
            ),
            cell=cell,
            pbc=torch.tensor([True, True, True]),
        )

        system = get_system_with_neighbor_lists(
            system, model.requested_neighbor_lists()
        )

        outputs = {"energy": ModelOutput(per_atom=False)}
        output = model([system], outputs)
        energy = output["energy"].block().values.sum()
        return energy

    assert torch.autograd.gradcheck(compute, cell, fast_mode=True)
    assert torch.autograd.gradgradcheck(compute, cell, fast_mode=True)

    if torch.cuda.is_available():
        cell_cuda = cell.detach().cuda().requires_grad_(True)
        assert torch.autograd.gradcheck(compute, cell_cuda, fast_mode=True)
        assert torch.autograd.gradgradcheck(compute, cell_cuda, fast_mode=True)



import json
from typing import Any, Union

import torch
from omegaconf import DictConfig, ListConfig, OmegaConf
from omegaconf.basecontainer import BaseContainer

from .. import PACKAGE_ROOT, RANDOM_SEED
from .architectures import import_architecture
from .devices import pick_devices
from .jsonschema import validate


def _get_architecture_model(conf: BaseContainer) -> Any:
    architecture = import_architecture(conf["architecture"]["name"])
    return architecture.__model__


def default_device(_root_: BaseContainer) -> str:
    """Custom OmegaConf resolver to find the default device of an architecture.

    Device is found using the :py:func:metatrain.utils.devices.pick_devices`
    function."""

    Model = _get_architecture_model(_root_)
    desired_device = pick_devices(Model.__supported_devices__)

    if len(desired_device) > 1:
        return "multi-cuda"
    else:
        return desired_device[0].type


def default_precision(_root_: BaseContainer) -> int:
    """Custom OmegaConf resolver to find the default precision of an architecture.

    File format is obtained based on the architecture name and its first entry in the
    ``supported_dtypes`` list."""

    Model = _get_architecture_model(_root_)

    # desired `dtype` is the first entry
    default_dtype = Model.__supported_dtypes__[0]

    # `base_precision` in options has to be a integer and not a torch.dtype
    if default_dtype in [torch.float64, torch.double]:
        return 64
    elif default_dtype == torch.float32:
        return 32
    elif default_dtype == torch.float16:
        return 16
    else:
        raise ValueError(
            f"architectures `default_dtype` ({default_dtype}) refers to an unknown "
            "torch dtype. This should not happen."
        )


def default_random_seed() -> int:
    """Return session seed in the range [0, 2**32)."""
    return RANDOM_SEED


# Register custom resolvers
OmegaConf.register_new_resolver("default_device", default_device)
OmegaConf.register_new_resolver("default_precision", default_precision)
OmegaConf.register_new_resolver("default_random_seed", default_random_seed)


def _resolve_single_str(config: str) -> DictConfig:
    return OmegaConf.create({"read_from": config})


# Base options/configurations
BASE_OPTIONS = OmegaConf.create(
    {
        "device": "${default_device:}",
        "base_precision": "${default_precision:}",
        "seed": "${default_random_seed:}",
    }
)


CONF_SYSTEMS = OmegaConf.create(
    {
        "read_from": "${..read_from}",
        "reader": None,
        "length_unit": None,
    }
)

CONF_TARGET_FIELDS = OmegaConf.create(
    {
        "quantity": "energy",
        "read_from": "${...systems.read_from}",
        "reader": None,
        "key": None,
        "unit": None,
        "per_atom": False,
        "type": "scalar",
        "num_subtargets": 1,
    }
)

CONF_GRADIENTS = OmegaConf.create({"forces": False, "stress": False, "virial": False})
CONF_GRADIENT = OmegaConf.create(
    {
        "read_from": "${..read_from}",
        "reader": None,
        "key": None,
    }
)

KNOWN_GRADIENTS = list(CONF_GRADIENTS.keys())

# Merge configs to get default configs for energies and other targets
CONF_TARGET = OmegaConf.merge(CONF_TARGET_FIELDS, CONF_GRADIENTS)
CONF_ENERGY = CONF_TARGET.copy()
CONF_ENERGY["forces"] = CONF_GRADIENT.copy()
CONF_ENERGY["stress"] = CONF_GRADIENT.copy()

# Schema with the dataset options
with open(PACKAGE_ROOT / "share/schema-dataset.json") as f:
    SCHEMA_DATASET = json.load(f)


def check_dataset_options(dataset_config: ListConfig) -> None:
    """Perform consistency checks within one dataset config.

    This is useful if the dataset config is made of several datasets.

    - The function checks if ``length_units`` in each system section are known and the
       same.
    - For unknown quantities a warning is given.
    - If the names of the ``"targets"`` sections are the same between the elements of
       the list of datasets also the units must be the same.
    - Two targets with the names `{target}` and `mtt::{target}` are not allowed.

    :param dataset_config: A List of configuration to be checked. In the list contains
        only one element no checks are performed.
    :raises ValueError: If the units are not consistent between the dataset options or
        if two different targets have the `{target}` and `mtt::{target}` names.
    """
    desired_config = dataset_config[0]

    if hasattr(desired_config, "targets"):
        # save unit for each target seaction for later comparison
        unit_dict = {k: v["unit"] for k, v in desired_config["targets"].items()}
    else:
        unit_dict = {}

    if hasattr(desired_config, "systems"):
        desired_length_unit = desired_config["systems"]["length_unit"]
    else:
        desired_length_unit = None

    # loop over ALL configs because we have check units for all elements in
    # `dataset_config`
    for actual_config in dataset_config:
        if desired_length_unit:
            # Perform consistency checks between config elements
            actual_length_unit = actual_config["systems"]["length_unit"]
            if actual_length_unit != desired_length_unit:
                raise ValueError(
                    "`length_unit`s are inconsistent between one of the dataset "
                    f"options. {actual_length_unit!r} != {desired_length_unit!r}."
                )

        if hasattr(actual_config, "targets"):
            for target_key, target in actual_config["targets"].items():
                unit = target["unit"]

                # If a target section name is not part of the saved units we add it for
                # later comparison. We do not have to start the loop again because this
                # target section name is not present in one of the datasets checked
                # before.
                if target_key not in unit_dict.keys():
                    unit_dict[target_key] = unit

                if unit_dict[target_key] != unit:
                    raise ValueError(
                        f"Units of target section {target_key!r} are inconsistent. "
                        f"Found {unit!r} and {unit_dict[target_key]!r}!"
                    )

        # `target` and `mtt::target` are not allowed to be present at the same time
        if hasattr(actual_config, "targets"):
            for target_key in actual_config["targets"].keys():
                if f"mtt::{target_key}" in actual_config["targets"].keys():
                    raise ValueError(
                        f"Two targets with the names `{target_key}` and "
                        f"`mtt::{target_key}` are not allowed to be present "
                        "at the same time."
                    )


def expand_dataset_config(conf: Union[str, DictConfig, ListConfig]) -> ListConfig:
    """Expands shorthand notations in a dataset configuration to its full format.

    This function takes a dataset configuration, either as a :py:class:str,
    :py:class:`omegaconf.DictConfig` or a :py:class:`omegaconf.ListConfig`, and expands
    it into a detailed configuration format. It processes systems, targets, and gradient
    sections, setting default values and inferring missing information. Unknown keys are
    ignored, allowing for flexibility.

    If the dataset configuration is either a :class:`str` or a
    :class:`omegaconf.DictConfig`

    The function performs the following steps for each config

    - Merges and interpolates the input configuration with the base configurations.
    - Expands shorthand notations like file paths or simple true/false settings to full
      dictionary systems. This includes setting the units to the base units of
      ``"angstrom"`` and ``"eV"``.
    - Handles special cases, such as the mandatory nature of the "energy" section for MD
      simulations and the mutual exclusivity of 'stress' and 'virial' sections.
      Additionally the gradient sections for "forces" are enables by default.

    :param conf: The dataset configuration, either as a file path string or a DictConfig
        object.
    :raises ValueError: If both ``virial`` and ``stress`` sections are enabled in the
        "energy" target, as this is not permissible for training.
    :returns: List of datasets configurations. If ``conf`` was a :class:`str` or a
        :class:`omegaconf.DictConfig` the list contains only a single element.
    """
    # Expand str -> DictConfig
    if isinstance(conf, str):
        read_from = conf
        conf = OmegaConf.create(
            {"systems": read_from, "targets": {"energy": read_from}}
        )

    # Expand DictConfig -> ListConfig
    if isinstance(conf, DictConfig):
        conf = OmegaConf.create([conf])

    # Perform expansion per config inside the ListConfig
    for conf_element in conf:
        validate(instance=OmegaConf.to_container(conf_element), schema=SCHEMA_DATASET)
        if hasattr(conf_element, "systems"):
            if type(conf_element["systems"]) is str:
                conf_element["systems"] = _resolve_single_str(conf_element["systems"])

            conf_element["systems"] = OmegaConf.merge(
                CONF_SYSTEMS, conf_element["systems"]
            )

        if hasattr(conf_element, "targets"):
            for target_key, target in conf_element["targets"].items():
                if type(target) is str:
                    target = _resolve_single_str(target)

                # for special case "energy" we enable sections for `forces` and `stress`
                # gradients by default
                if target_key == "energy":
                    target = OmegaConf.merge(CONF_ENERGY, target)
                else:
                    target = OmegaConf.merge(CONF_TARGET, target)

                if target["key"] is None:
                    target["key"] = target_key

                # update DictConfig to allow for config node interpolation
                conf_element["targets"][target_key] = target

                # merge and interpolate possibly present gradients with default gradient
                # config
                for gradient_key, gradient_conf in conf_element["targets"][
                    target_key
                ].items():
                    if gradient_key in KNOWN_GRADIENTS:
                        if gradient_conf is True:
                            gradient_conf = CONF_GRADIENT.copy()
                        elif type(gradient_conf) is str:
                            gradient_conf = _resolve_single_str(gradient_conf)

                        if isinstance(gradient_conf, DictConfig):
                            gradient_conf = OmegaConf.merge(
                                CONF_GRADIENT, gradient_conf
                            )

                            if gradient_conf["key"] is None:
                                gradient_conf["key"] = gradient_key

                            conf_element["targets"][target_key][gradient_key] = (
                                gradient_conf
                            )

                # If user sets the virial gradient and leaves the stress gradient
                # untouched, we disable the by default enabled stress gradient section.
                base_stress_gradient_conf = CONF_GRADIENT.copy()
                base_stress_gradient_conf["key"] = "stress"

                if (
                    target_key == "energy"
                    and conf_element["targets"][target_key]["virial"]
                    and conf_element["targets"][target_key]["stress"]
                    == base_stress_gradient_conf
                ):
                    conf_element["targets"][target_key]["stress"] = False

                if (
                    conf_element["targets"][target_key]["stress"]
                    and conf_element["targets"][target_key]["virial"]
                ):
                    raise ValueError(
                        f"Cannot perform training with respect to virials and stress "
                        f"as in section {target_key}. Set either `virials: off` or "
                        "`stress: off`."
                    )

    check_dataset_options(conf)
    return conf


def check_units(
    actual_options: Union[DictConfig, ListConfig],
    desired_options: Union[DictConfig, ListConfig],
) -> None:
    """Perform consistency checks between two dataset configs.

    :param actual_options: The dataset options that you want to test.
    :param desired_options: The dataset options ``actual_options`` is tested against.

    :raises ValueError: If the length units are not consistent between
        the system in the dataset options.
    :raises ValueError: If a target is present only in desider_option and
        not in actual_option.
    :raises ValueError: If the unit of a target quantity is not consistent between
        the dataset option.
    """
    if type(actual_options) is DictConfig:
        actual_options = OmegaConf.create([actual_options])
    if type(desired_options) is DictConfig:
        desired_options = OmegaConf.create([desired_options])

    if len(actual_options) != len(desired_options):
        raise ValueError(
            f"Length of actual_options ({len(actual_options)}) and desired_options "
            f"({len(desired_options)}) is different!"
        )

    for actual_options_element, desired_options_element in zip(
        actual_options,
        desired_options,
    ):
        actual_length_unit = actual_options_element["systems"]["length_unit"]
        desired_length_unit = desired_options_element["systems"]["length_unit"]

        if actual_length_unit != desired_length_unit:
            raise ValueError(
                "`length_unit`s are inconsistent between one of the dataset options. "
                f"{actual_length_unit!r} != {desired_length_unit!r}."
            )

        for target in actual_options_element["targets"]:
            actual_unit = actual_options_element["targets"][target]["unit"]
            if target in desired_options_element["targets"]:
                desired_unit = desired_options_element["targets"][target]["unit"]
                if actual_unit != desired_unit:
                    raise ValueError(
                        f"Units of target {target!r} are inconsistent between one of "
                        f"the dataset options. {actual_unit!r} != {desired_unit!r}."
                    )
            else:
                raise ValueError(
                    f"Target {target!r} is not present in one of the given dataset "
                    "options."
                )



import re
import warnings

import yaml  # type: ignore


def propagate_duplicated_params(provided_hypers, default_hypers, first_key, second_key):
    if (first_key in provided_hypers.keys()) and (second_key in provided_hypers.keys()):
        raise ValueError(f"only one of {first_key} and {second_key} should be provided")

    if (first_key in default_hypers.keys()) and (second_key in default_hypers.keys()):
        raise ValueError(
            f"only one of {first_key} and {second_key} should be in default hypers"
        )

    output_key, output_value = None, None
    for key in [first_key, second_key]:
        if key in provided_hypers.keys():
            output_key = key
            output_value = provided_hypers[key]

    if output_key is None:
        for key in [first_key, second_key]:
            if key in default_hypers.keys():
                output_key = key
                output_value = default_hypers[key]

    if output_key is None:
        raise ValueError(f"{first_key} or {second_key} must be provided somewhere")

    return output_key, output_value


def check_is_shallow(hypers):
    for key in hypers.keys():
        if isinstance(hypers[key], dict):
            raise ValueError("Nesting of more than two is not supported")


def combine_hypers(provided_hypers, default_hypers):
    group_keys = [
        "ARCHITECTURAL_HYPERS",
        "FITTING_SCHEME",
        "MLIP_SETTINGS",
        "GENERAL_TARGET_SETTINGS",
        "UTILITY_FLAGS",
    ]

    for key in provided_hypers.keys():
        if key not in group_keys:
            raise ValueError(f"unknown hyper parameter {key}")

    for key in default_hypers.keys():
        if key not in group_keys:
            raise ValueError(f"unknown hyper parameter {key}")

    result = {}
    for key in group_keys:
        default_now = default_hypers[key]
        if key in provided_hypers.keys():
            provided_now = provided_hypers[key]
        else:
            provided_now = {}
        if key == "FITTING_SCHEME":
            duplicated_params = [
                ["ATOMIC_BATCH_SIZE", "STRUCTURAL_BATCH_SIZE"],
                ["EPOCH_NUM", "EPOCH_NUM_ATOMIC"],
                ["SCHEDULER_STEP_SIZE", "SCHEDULER_STEP_SIZE_ATOMIC"],
                ["EPOCHS_WARMUP", "EPOCHS_WARMUP_ATOMIC"],
            ]
        else:
            duplicated_params = []
        result[key] = combine_hypers_shallow(
            provided_now, default_now, duplicated_params
        )

    if (not result["MLIP_SETTINGS"]["USE_ENERGIES"]) and (
        not result["MLIP_SETTINGS"]["USE_FORCES"]
    ):
        raise ValueError(
            "At least one of the energies and forces should be used for fitting"
        )

    if (not result["MLIP_SETTINGS"]["USE_ENERGIES"]) or (
        not result["MLIP_SETTINGS"]["USE_FORCES"]
    ):
        if result["FITTING_SCHEME"]["ENERGY_WEIGHT"] is not None:
            warnings.warn(
                "ENERGY_WEIGHT was provided, but in the current calculation, "
                + "it doesn't ffect anything since only one target of energies and "
                + "forces is used",
                stacklevel=2,
            )

    if result["ARCHITECTURAL_HYPERS"]["USE_ADDITIONAL_SCALAR_ATTRIBUTES"]:
        if result["ARCHITECTURAL_HYPERS"]["SCALAR_ATTRIBUTES_SIZE"] is None:
            raise ValueError(
                "scalar attributes size must be provided if "
                "use_additional_scalar_attributes == True"
            )

    if result["FITTING_SCHEME"]["DO_GRADIENT_CLIPPING"]:
        if result["FITTING_SCHEME"]["GRADIENT_CLIPPING_MAX_NORM"] is None:
            raise ValueError(
                "gradient clipping max_norm must be provided if "
                "do_gradient_clipping == True"
            )

    if result["FITTING_SCHEME"]["BALANCED_DATA_LOADER"]:
        if "STRUCTURAL_BATCH_SIZE" in result["FITTING_SCHEME"].keys():
            if result["FITTING_SCHEME"]["STRUCTURAL_BATCH_SIZE"] is not None:
                raise ValueError(
                    "if using balanced_data_loader only atomic batch size "
                    "can be provided"
                )

    if result["FITTING_SCHEME"]["MODEL_TO_START_WITH"] is not None:
        if result["FITTING_SCHEME"]["ALL_SPECIES_PATH"] is None:
            raise ValueError(
                "When fine-tuning the model starting from a checkpoint "
                "[MODEL_TO_START_WITH, the path to the all_species.npy "
                " file [ALL_SPECIES_PATH] must be provided."
            )

    return result


def combine_hypers_shallow(provided_hypers, default_hypers, duplicated_params):
    check_is_shallow(provided_hypers)
    check_is_shallow(default_hypers)

    duplicated_params_unrolled = []
    for el in duplicated_params:
        duplicated_params_unrolled.append(el[0])
        duplicated_params_unrolled.append(el[1])

    for key in provided_hypers.keys():
        if key not in default_hypers.keys():
            if key not in duplicated_params_unrolled:
                raise ValueError(f"unknown hyper parameter {key}")

    result = {}

    for key in default_hypers.keys():
        if key in provided_hypers.keys():
            if key not in duplicated_params_unrolled:
                result[key] = provided_hypers[key]
        else:
            if key not in duplicated_params_unrolled:
                result[key] = default_hypers[key]

    for el in duplicated_params:
        dupl_key, dupl_value = propagate_duplicated_params(
            provided_hypers, default_hypers, el[0], el[1]
        )
        result[dupl_key] = dupl_value

    return result


def fix_Nones_in_yaml(hypers_dict):
    for key in hypers_dict.keys():
        if (hypers_dict[key] == "None") or (hypers_dict[key] == "none"):
            hypers_dict[key] = None
        if isinstance(hypers_dict[key], dict):
            fix_Nones_in_yaml(hypers_dict[key])


class Hypers:
    def __init__(self, hypers_dict):
        for key, value in hypers_dict.items():
            if isinstance(value, dict):
                self.__dict__[key] = Hypers(value)
            else:
                self.__dict__[key] = value


def load_hypers_from_file(path_to_hypers):
    loader = yaml.SafeLoader
    loader.add_implicit_resolver(
        "tag:yaml.org,2002:float",
        re.compile(
            """^(?:
            [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
        |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
        |\\.[0-9_]+(?:[eE][-+][0-9]+)?
        |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
        |[-+]?\\.(?:inf|Inf|INF)
        |\\.(?:nan|NaN|NAN))$""",
            re.X,
        ),
        list("-+0123456789."),
    )

    with open(path_to_hypers, "r") as f:
        hypers = yaml.load(f, Loader=loader)
        fix_Nones_in_yaml(hypers)

    return Hypers(hypers)


def set_hypers_from_files(path_to_provided_hypers, path_to_default_hypers):
    loader = yaml.SafeLoader
    loader.add_implicit_resolver(
        "tag:yaml.org,2002:float",
        re.compile(
            """^(?:
            [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
        |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
        |\\.[0-9_]+(?:[eE][-+][0-9]+)?
        |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
        |[-+]?\\.(?:inf|Inf|INF)
        |\\.(?:nan|NaN|NAN))$""",
            re.X,
        ),
        list("-+0123456789."),
    )

    with open(path_to_provided_hypers, "r") as f:
        provided_hypers = yaml.load(f, Loader=loader)
        fix_Nones_in_yaml(provided_hypers)

    with open(path_to_default_hypers, "r") as f:
        default_hypers = yaml.load(f, Loader=loader)
        fix_Nones_in_yaml(default_hypers)

    combined_hypers = combine_hypers(provided_hypers, default_hypers)
    return Hypers(combined_hypers)


def hypers_to_dict(obj):
    if isinstance(obj, Hypers):
        return {key: hypers_to_dict(value) for key, value in obj.__dict__.items()}
    else:
        return obj


def save_hypers(hypers, path_save):
    hypers_dict = hypers_to_dict(hypers)
    with open(path_save, "w") as f:
        yaml.dump(hypers_dict, f)



import random

import numpy as np
import torch
from metatensor.torch.atomistic import ModelOutput
from omegaconf import OmegaConf

from metatrain.experimental.nativepet import NativePET, Trainer
from metatrain.utils.data import Dataset, DatasetInfo
from metatrain.utils.data.readers import (
    read_systems,
    read_targets,
)
from metatrain.utils.data.target_info import get_energy_target_info
from metatrain.utils.evaluate_model import evaluate_model
from metatrain.utils.neighbor_lists import get_system_with_neighbor_lists

from . import DATASET_PATH, DATASET_WITH_FORCES_PATH, DEFAULT_HYPERS, MODEL_HYPERS


def test_regression_init():
    """Perform a regression test on the model at initialization"""

    # reproducibility
    random.seed(0)
    np.random.seed(0)
    torch.manual_seed(0)

    targets = {}
    targets["mtt::U0"] = get_energy_target_info({"quantity": "energy", "unit": "eV"})

    dataset_info = DatasetInfo(
        length_unit="Angstrom", atomic_types=[1, 6, 7, 8], targets=targets
    )
    model = NativePET(MODEL_HYPERS, dataset_info)

    # Predict on the first five systems
    systems = read_systems(DATASET_PATH)[:5]
    systems = [system.to(torch.float32) for system in systems]
    for system in systems:
        get_system_with_neighbor_lists(system, model.requested_neighbor_lists())

    output = model(
        systems,
        {"mtt::U0": ModelOutput(quantity="energy", unit="", per_atom=False)},
    )

    expected_output = torch.tensor(
        [
            [3.015289306641],
            [1.845376491547],
            [0.776397109032],
            [1.858697414398],
            [1.014655113220],
        ]
    )

    # if you need to change the hardcoded values:
    torch.set_printoptions(precision=12)
    print(output["mtt::U0"].block().values)

    torch.testing.assert_close(output["mtt::U0"].block().values, expected_output)


def test_regression_energies_train():
    """Perform a regression test on the model when
    trained for 2 epoch on a small dataset, energies only"""

    # reproducibility
    random.seed(0)
    np.random.seed(0)
    torch.manual_seed(0)

    systems = read_systems(DATASET_PATH)

    conf = {
        "mtt::U0": {
            "quantity": "energy",
            "read_from": DATASET_PATH,
            "reader": "ase",
            "key": "U0",
            "unit": "eV",
            "type": "scalar",
            "per_atom": False,
            "num_subtargets": 1,
            "forces": False,
            "stress": False,
            "virial": False,
        }
    }
    targets, target_info_dict = read_targets(OmegaConf.create(conf))
    dataset = Dataset.from_dict({"system": systems, "mtt::U0": targets["mtt::U0"]})

    hypers = DEFAULT_HYPERS.copy()
    hypers["training"]["num_epochs"] = 2
    hypers["training"]["scheduler_patience"] = 1

    dataset_info = DatasetInfo(
        length_unit="Angstrom", atomic_types=[1, 6, 7, 8], targets=target_info_dict
    )
    model = NativePET(MODEL_HYPERS, dataset_info)

    trainer = Trainer(hypers["training"])
    trainer.train(
        model=model,
        dtype=torch.float32,
        devices=[torch.device("cpu")],
        train_datasets=[dataset],
        val_datasets=[dataset],
        checkpoint_dir=".",
    )

    # Predict on the first five systems
    systems = [system.to(torch.float32) for system in systems]
    for system in systems:
        get_system_with_neighbor_lists(system, model.requested_neighbor_lists())
    output = model(
        systems[:5],
        {"mtt::U0": ModelOutput(quantity="energy", unit="", per_atom=False)},
    )

    expected_output = torch.tensor(
        [
            [3.012795209885],
            [1.843898773193],
            [0.775631427765],
            [1.857149720192],
            [1.013873577118],
        ]
    )

    # if you need to change the hardcoded values:
    torch.set_printoptions(precision=12)
    print(output["mtt::U0"].block().values)

    torch.testing.assert_close(output["mtt::U0"].block().values, expected_output)


def test_regression_energies_forces_train():
    """Perform a regression test on the model when
    trained for 2 epoch on a small dataset with energies
    and forces"""

    # reproducibility
    random.seed(0)
    np.random.seed(0)
    torch.manual_seed(0)

    systems = read_systems(DATASET_WITH_FORCES_PATH)

    conf = {
        "energy": {
            "quantity": "energy",
            "read_from": DATASET_WITH_FORCES_PATH,
            "reader": "ase",
            "key": "energy",
            "unit": "eV",
            "type": "scalar",
            "per_atom": False,
            "num_subtargets": 1,
            "forces": {"read_from": DATASET_WITH_FORCES_PATH, "key": "force"},
            "stress": False,
            "virial": False,
        }
    }

    targets, target_info_dict = read_targets(OmegaConf.create(conf))
    targets = {"energy": targets["energy"]}
    dataset = Dataset.from_dict({"system": systems, "energy": targets["energy"]})
    hypers = DEFAULT_HYPERS.copy()
    hypers["training"]["num_epochs"] = 2
    hypers["training"]["scheduler_patience"] = 1
    hypers["training"]["fixed_composition_weights"] = {}

    dataset_info = DatasetInfo(
        length_unit="Angstrom", atomic_types=[6], targets=target_info_dict
    )
    model = NativePET(MODEL_HYPERS, dataset_info)
    trainer = Trainer(hypers["training"])
    trainer.train(
        model=model,
        dtype=torch.float32,
        devices=[torch.device("cpu")],
        train_datasets=[dataset],
        val_datasets=[dataset],
        checkpoint_dir=".",
    )

    # Predict on the first five systems
    systems = [system.to(torch.float32) for system in systems]
    for system in systems:
        get_system_with_neighbor_lists(system, model.requested_neighbor_lists())

    output = evaluate_model(
        model, systems[:5], targets=target_info_dict, is_training=False
    )

    expected_output = torch.tensor(
        [
            [20.386034011841],
            [20.353490829468],
            [20.303865432739],
            [20.413286209106],
            [20.318788528442],
        ]
    )

    expected_gradients_output = torch.tensor(
        [0.208536088467, -0.117365449667, -0.278660595417]
    )

    # if you need to change the hardcoded values:
    torch.set_printoptions(precision=12)
    print(output["energy"].block().values)
    print(output["energy"].block().gradient("positions").values.squeeze(-1)[0])

    torch.testing.assert_close(output["energy"].block().values, expected_output)
    torch.testing.assert_close(
        output["energy"].block().gradient("positions").values[0, :, 0],
        expected_gradients_output,
    )



from metatensor.learn.data import DataLoader

from ...utils.additive import remove_additive
from ...utils.data import collate_fn
from ...utils.data.system_to_ase import system_to_ase
from ...utils.neighbor_lists import (
    get_requested_neighbor_lists,
    get_system_with_neighbor_lists,
)


# dummy dataloaders due to https://github.com/metatensor/metatensor/issues/521
def dataset_to_ase(dataset, model, do_forces=True, target_name="energy"):
    dataloader = DataLoader(
        dataset,
        batch_size=1,
        shuffle=False,
        collate_fn=collate_fn,
    )
    ase_dataset = []
    for (system,), targets in dataloader:
        # remove additive model (e.g. ZBL) contributions
        requested_neighbor_lists = get_requested_neighbor_lists(model)
        system = get_system_with_neighbor_lists(system, requested_neighbor_lists)
        for additive_model in model.additive_models:
            targets = remove_additive(
                [system], targets, additive_model, model.dataset_info.targets
            )
        # transform to ase atoms
        ase_atoms = system_to_ase(system)
        ase_atoms.info["energy"] = float(
            targets[target_name].block().values.squeeze(-1).detach().cpu().numpy()
        )
        if do_forces:
            ase_atoms.arrays["forces"] = (
                -targets[target_name]
                .block()
                .gradient("positions")
                .values.squeeze(-1)
                .detach()
                .cpu()
                .numpy()
            )
        ase_dataset.append(ase_atoms)
    return ase_dataset



import datetime
import logging
import os
import pickle
import time
import warnings
from pathlib import Path
from typing import List, Union

import numpy as np
import torch
from torch_geometric.nn import DataParallel

from ..utils.data import Dataset, check_datasets
from ..utils.io import check_file_extension
from . import PET as WrappedPET
from .modules.analysis import adapt_hypers
from .modules.data_preparation import (
    get_corrected_energies,
    get_forces,
    get_pyg_graphs,
    get_self_contributions,
    update_pyg_graphs,
)
from .modules.hypers import Hypers, save_hypers
from .modules.pet import PET, FlagsWrapper, PETMLIPWrapper, PETUtilityWrapper
from .modules.utilities import (
    FullLogger,
    ModelKeeper,
    dtype2string,
    get_calc_names,
    get_data_loaders,
    get_loss,
    get_optimizer,
    get_rmse,
    get_scheduler,
    log_epoch_stats,
    set_reproducibility,
    string2dtype,
)
from .utils import dataset_to_ase, load_raw_pet_model, update_hypers
from .utils.fine_tuning import LoRAWrapper


class Trainer:
    def __init__(self, train_hypers):
        self.hypers = {"FITTING_SCHEME": train_hypers}
        self.pet_dir = None
        self.pet_trainer_state = None
        self.epoch = None
        self.best_metric = None
        self.best_model_state_dict = None

    def train(
        self,
        model: WrappedPET,
        dtype: torch.dtype,
        devices: List[torch.device],
        train_datasets: List[Union[Dataset, torch.utils.data.Subset]],
        val_datasets: List[Union[Dataset, torch.utils.data.Subset]],
        checkpoint_dir: str,
    ):
        assert dtype in WrappedPET.__supported_dtypes__

        name_of_calculation = "pet"
        self.pet_dir = Path(checkpoint_dir) / name_of_calculation

        if len(train_datasets) != 1:
            raise ValueError("PET only supports a single training dataset")
        if len(val_datasets) != 1:
            raise ValueError("PET only supports a single validation dataset")
        if torch.device("cpu") in devices:
            warnings.warn(
                "Training PET on a CPU is very slow! For better performance, use a "
                "CUDA GPU.",
                stacklevel=1,
            )

        logging.info("Checking datasets for consistency")
        check_datasets(train_datasets, val_datasets)

        train_dataset = train_datasets[0]
        val_dataset = val_datasets[0]

        # are we fitting on only energies or energies and forces?
        target_name = model.target_name
        do_forces = (
            next(iter(train_dataset))[target_name].block().has_gradient("positions")
        )

        ase_train_dataset = dataset_to_ase(
            train_dataset, model, do_forces=do_forces, target_name=target_name
        )
        ase_val_dataset = dataset_to_ase(
            val_dataset, model, do_forces=do_forces, target_name=target_name
        )

        self.hypers = update_hypers(self.hypers, model.hypers, do_forces)

        device = devices[0]  # only one device, as we don't support multi-gpu for now

        ########################################
        # STARTING THE PURE PET TRAINING SCRIPT #
        ########################################

        logging.info("Initializing PET training...")

        TIME_SCRIPT_STARTED = time.time()
        value = datetime.datetime.fromtimestamp(TIME_SCRIPT_STARTED)
        logging.info(f"Starting training at: {value.strftime('%Y-%m-%d %H:%M:%S')}")
        training_configuration_log = "Training configuration:\n"
        training_configuration_log += f"Output directory: {checkpoint_dir}\n"
        training_configuration_log += f"Training using device: {device}\n"

        if not os.path.exists(checkpoint_dir):
            os.mkdir(checkpoint_dir)

        hypers = Hypers(self.hypers)
        dtype = string2dtype(hypers.ARCHITECTURAL_HYPERS.DTYPE)  # type: ignore
        torch.set_default_dtype(dtype)

        FITTING_SCHEME = hypers.FITTING_SCHEME  # type: ignore
        MLIP_SETTINGS = hypers.MLIP_SETTINGS  # type: ignore
        ARCHITECTURAL_HYPERS = hypers.ARCHITECTURAL_HYPERS  # type: ignore

        if model.is_lora_applied and not FITTING_SCHEME.USE_LORA_PEFT:
            raise ValueError(
                "LoRA is applied to the model, but the USE_LORA_PEFT is False"
                " in the training hyperparameters. Please set USE_LORA_PEFT to True"
            )

        if FITTING_SCHEME.USE_SHIFT_AGNOSTIC_LOSS:
            raise ValueError(
                "shift agnostic loss is intended only for general target training"
            )

        training_configuration_log += (
            f"Output dimensionality: {ARCHITECTURAL_HYPERS.D_OUTPUT}\n"
        )
        training_configuration_log += (
            f"Target type: {ARCHITECTURAL_HYPERS.TARGET_TYPE}\n"
        )
        training_configuration_log += (
            f"Target aggregation: {ARCHITECTURAL_HYPERS.TARGET_AGGREGATION}\n"
        )
        training_configuration_log += f"Random seed: {FITTING_SCHEME.RANDOM_SEED}\n"
        training_configuration_log += (
            f"CUDA is deterministic: {FITTING_SCHEME.CUDA_DETERMINISTIC}"
        )

        st = """
Legend: LR          -> Learning Rate
        MAE         -> Mean Absolute Error
        RMSE        -> Root Mean Square Error
        V-E-MAE/at  -> MAE of the Energy per atom on the Validation set
        V-E-RMSE/at -> RMSE of the Energy per atom on the Validation set
        V-F-MAE     -> MAE of the Forces on the Validation set
        V-F-RMSE    -> RMSE of the Forces on the Validation set
        T-E-MAE/at  -> MAE of the Energy per atom on the Training set
        T-E-RMSE/at -> RMSE of the Energy per atom on the Training set
        T-F-MAE     -> MAE of the Forces on the Training set
        T-F-RMSE    -> RMSE of the Forces on the Training set
Units of the Energy and Forces are the same units given in input"""
        training_configuration_log += st
        logging.info(training_configuration_log)

        set_reproducibility(
            FITTING_SCHEME.RANDOM_SEED, FITTING_SCHEME.CUDA_DETERMINISTIC
        )

        adapt_hypers(FITTING_SCHEME, ase_train_dataset)

        all_species = model.atomic_types

        name_to_load, NAME_OF_CALCULATION = get_calc_names(
            os.listdir(checkpoint_dir), name_of_calculation
        )

        os.mkdir(f"{checkpoint_dir}/{NAME_OF_CALCULATION}")
        np.save(f"{checkpoint_dir}/{NAME_OF_CALCULATION}/all_species.npy", all_species)
        hypers.UTILITY_FLAGS.CALCULATION_TYPE = "mlip"  # type: ignore
        save_hypers(hypers, f"{checkpoint_dir}/{NAME_OF_CALCULATION}/hypers_used.yaml")

        logging.info("Convering structures to PyG graphs...")

        train_graphs = get_pyg_graphs(
            ase_train_dataset,
            all_species,
            ARCHITECTURAL_HYPERS.R_CUT,
            ARCHITECTURAL_HYPERS.USE_ADDITIONAL_SCALAR_ATTRIBUTES,
            ARCHITECTURAL_HYPERS.USE_LONG_RANGE,
            ARCHITECTURAL_HYPERS.K_CUT,
            ARCHITECTURAL_HYPERS.N_TARGETS > 1,
            ARCHITECTURAL_HYPERS.TARGET_INDEX_KEY,
        )
        val_graphs = get_pyg_graphs(
            ase_val_dataset,
            all_species,
            ARCHITECTURAL_HYPERS.R_CUT,
            ARCHITECTURAL_HYPERS.USE_ADDITIONAL_SCALAR_ATTRIBUTES,
            ARCHITECTURAL_HYPERS.USE_LONG_RANGE,
            ARCHITECTURAL_HYPERS.K_CUT,
            ARCHITECTURAL_HYPERS.N_TARGETS > 1,
            ARCHITECTURAL_HYPERS.TARGET_INDEX_KEY,
        )

        logging.info("Pre-processing training data...")
        if MLIP_SETTINGS.USE_ENERGIES:
            if model.pet is not None:
                self_contributions = model.pet.self_contributions
            else:
                self_contributions = get_self_contributions(
                    MLIP_SETTINGS.ENERGY_KEY, ase_train_dataset, all_species
                )
            np.save(
                f"{checkpoint_dir}/{NAME_OF_CALCULATION}/self_contributions.npy",
                self_contributions,
            )

            train_energies = get_corrected_energies(
                MLIP_SETTINGS.ENERGY_KEY,
                ase_train_dataset,
                all_species,
                self_contributions,
            )
            val_energies = get_corrected_energies(
                MLIP_SETTINGS.ENERGY_KEY,
                ase_val_dataset,
                all_species,
                self_contributions,
            )

            update_pyg_graphs(train_graphs, "y", train_energies)
            update_pyg_graphs(val_graphs, "y", val_energies)

        if MLIP_SETTINGS.USE_FORCES:
            train_forces = get_forces(ase_train_dataset, MLIP_SETTINGS.FORCES_KEY)
            val_forces = get_forces(ase_val_dataset, MLIP_SETTINGS.FORCES_KEY)

            update_pyg_graphs(train_graphs, "forces", train_forces)
            update_pyg_graphs(val_graphs, "forces", val_forces)

        train_loader, val_loader = get_data_loaders(
            train_graphs, val_graphs, FITTING_SCHEME
        )

        logging.info("Initializing the model...")
        if model.pet is not None:
            pet_model = model.pet.model
            if model.is_lora_applied:
                pet_model.model.hypers.TARGET_TYPE = "structural"
                pet_model.model.TARGET_TYPE = "structural"
            else:
                pet_model.hypers.TARGET_TYPE = "structural"
                pet_model.TARGET_TYPE = "structural"
        else:
            pet_model = PET(ARCHITECTURAL_HYPERS, 0.0, len(all_species))
        num_params = sum([p.numel() for p in pet_model.parameters()])
        logging.info(f"Number of parameters: {num_params}")

        if FITTING_SCHEME.USE_LORA_PEFT:
            if not model.is_lora_applied:
                lora_rank = FITTING_SCHEME.LORA_RANK
                lora_alpha = FITTING_SCHEME.LORA_ALPHA
                pet_model = LoRAWrapper(pet_model, lora_rank, lora_alpha)
                model.is_lora_applied = True

            num_trainable_params = sum(
                [p.numel() for p in pet_model.parameters() if p.requires_grad]
            )
            fraction = num_trainable_params / num_params * 100
            logging.info(
                f"Using LoRA PEFT with rank {FITTING_SCHEME.LORA_RANK} "
                + f"and alpha {FITTING_SCHEME.LORA_ALPHA}"
            )
            logging.info(
                "Number of trainable parameters: "
                + f"{num_trainable_params} [{fraction:.2f}%]"
            )
        pet_model = pet_model.to(device=device, dtype=dtype)
        pet_model = PETUtilityWrapper(pet_model, FITTING_SCHEME.GLOBAL_AUG)

        pet_model = PETMLIPWrapper(
            pet_model, MLIP_SETTINGS.USE_ENERGIES, MLIP_SETTINGS.USE_FORCES
        )
        if FITTING_SCHEME.MULTI_GPU and torch.cuda.is_available():
            logging.info(
                f"Using multi-GPU training on {torch.cuda.device_count()} GPUs"
            )
            pet_model = DataParallel(FlagsWrapper(pet_model))
            pet_model = pet_model.to(torch.device("cuda:0"))

        optim = get_optimizer(pet_model, FITTING_SCHEME)
        scheduler = get_scheduler(optim, FITTING_SCHEME)

        if self.pet_trainer_state is not None:
            for i, param_group in enumerate(optim.param_groups):
                if len(param_group["params"]) != len(
                    self.pet_trainer_state["optim_state_dict"]["param_groups"][i][
                        "params"
                    ]
                ):
                    raise RuntimeError(
                        "The number of parameters in the optimizer state dict "
                        "from the loaded checkpoint does not match the current "
                        "optimizer state. This means the model architecture has "
                        "changed since the last checkpoint. If you are using LoRA "
                        "PEFT, you should use the best model chekpoint from the "
                        "pre-training step. If you still need to use the current "
                        "checkpoint, set the trainer_state_dict in the checkpoint "
                        "to None and restart the training."
                    )

            optim.load_state_dict(self.pet_trainer_state["optim_state_dict"])
            scheduler.load_state_dict(self.pet_trainer_state["scheduler_state_dict"])
        else:
            logging.info(
                "No optimizer and scheduler state found in the "
                "checkpoint, starting from scratch"
            )

        history = []
        if MLIP_SETTINGS.USE_ENERGIES:
            energies_logger = FullLogger(
                FITTING_SCHEME.SUPPORT_MISSING_VALUES,
                FITTING_SCHEME.USE_SHIFT_AGNOSTIC_LOSS,
                device,
            )

        if MLIP_SETTINGS.USE_FORCES:
            forces_logger = FullLogger(
                FITTING_SCHEME.SUPPORT_MISSING_VALUES,
                FITTING_SCHEME.USE_SHIFT_AGNOSTIC_LOSS,
                device,
            )

        if MLIP_SETTINGS.USE_FORCES:
            val_forces = torch.cat(val_forces, dim=0)

            sliding_forces_rmse = get_rmse(
                val_forces.data.cpu().to(dtype=torch.float32).numpy(), 0.0
            )

            forces_rmse_model_keeper = ModelKeeper()
            forces_mae_model_keeper = ModelKeeper()

        if MLIP_SETTINGS.USE_ENERGIES:
            if FITTING_SCHEME.ENERGIES_LOSS == "per_structure":
                sliding_energies_rmse = get_rmse(val_energies, np.mean(val_energies))
            else:
                val_n_atoms = np.array(
                    [len(struc.positions) for struc in ase_val_dataset]
                )
                val_energies_per_atom = val_energies / val_n_atoms
                sliding_energies_rmse = get_rmse(
                    val_energies_per_atom, np.mean(val_energies_per_atom)
                )

            energies_rmse_model_keeper = ModelKeeper()
            energies_mae_model_keeper = ModelKeeper()

        if MLIP_SETTINGS.USE_ENERGIES and MLIP_SETTINGS.USE_FORCES:
            multiplication_rmse_model_keeper = ModelKeeper()
            multiplication_mae_model_keeper = ModelKeeper()

        if FITTING_SCHEME.EPOCHS_WARMUP > 0:
            remaining_lr_scheduler_steps = max(
                FITTING_SCHEME.EPOCHS_WARMUP - scheduler.last_epoch, 0
            )
            if remaining_lr_scheduler_steps > 0:
                lr_sheduler_msg = (
                    f" with {remaining_lr_scheduler_steps} steps of LR warmup"
                )
            else:
                lr_sheduler_msg = ""
        else:
            lr_sheduler_msg = ""
        logging.info(
            f"Starting training for {FITTING_SCHEME.EPOCH_NUM} epochs" + lr_sheduler_msg
        )
        TIME_TRAINING_STARTED = time.time()
        last_elapsed_time = 0
        if self.best_metric is None:
            self.best_metric = float("inf")
        start_epoch = 1 if self.epoch is None else self.epoch + 1
        for epoch in range(start_epoch, start_epoch + FITTING_SCHEME.EPOCH_NUM):
            pet_model.train(True)
            for batch in train_loader:
                if not FITTING_SCHEME.MULTI_GPU:
                    batch.to(device)

                if FITTING_SCHEME.MULTI_GPU:
                    pet_model.module.augmentation = True
                    pet_model.module.create_graph = True
                    predictions_energies, predictions_forces = pet_model(batch)
                else:
                    predictions_energies, predictions_forces = pet_model(
                        batch, augmentation=True, create_graph=True
                    )

                if FITTING_SCHEME.MULTI_GPU:
                    y_list = [el.y for el in batch]
                    batch_y = torch.tensor(
                        y_list, dtype=torch.get_default_dtype(), device=device
                    )

                    n_atoms_list = [el.n_atoms for el in batch]
                    batch_n_atoms = torch.tensor(
                        n_atoms_list, dtype=torch.get_default_dtype(), device=device
                    )

                else:
                    batch_y = batch.y
                    batch_n_atoms = batch.n_atoms

                if FITTING_SCHEME.ENERGIES_LOSS == "per_atom":
                    predictions_energies = predictions_energies / batch_n_atoms
                    ground_truth_energies = batch_y / batch_n_atoms
                else:
                    ground_truth_energies = batch_y

                if MLIP_SETTINGS.USE_ENERGIES:
                    energies_logger.train_logger.update(
                        predictions_energies, ground_truth_energies
                    )
                    loss_energies = get_loss(
                        predictions_energies,
                        ground_truth_energies,
                        FITTING_SCHEME.SUPPORT_MISSING_VALUES,
                        FITTING_SCHEME.USE_SHIFT_AGNOSTIC_LOSS,
                    )
                if MLIP_SETTINGS.USE_FORCES:
                    if FITTING_SCHEME.MULTI_GPU:
                        forces_list = [el.forces for el in batch]
                        batch_forces = torch.cat(forces_list, dim=0).to(device)
                    else:
                        batch_forces = batch.forces

                    forces_logger.train_logger.update(predictions_forces, batch_forces)
                    loss_forces = get_loss(
                        predictions_forces,
                        batch_forces,
                        FITTING_SCHEME.SUPPORT_MISSING_VALUES,
                        FITTING_SCHEME.USE_SHIFT_AGNOSTIC_LOSS,
                    )
                if MLIP_SETTINGS.USE_ENERGIES and MLIP_SETTINGS.USE_FORCES:
                    loss = FITTING_SCHEME.ENERGY_WEIGHT * loss_energies / (
                        sliding_energies_rmse**2
                    ) + loss_forces / (sliding_forces_rmse**2)
                    loss.backward()

                if MLIP_SETTINGS.USE_ENERGIES and (not MLIP_SETTINGS.USE_FORCES):
                    loss_energies.backward()
                if MLIP_SETTINGS.USE_FORCES and (not MLIP_SETTINGS.USE_ENERGIES):
                    loss_forces.backward()

                if FITTING_SCHEME.DO_GRADIENT_CLIPPING:
                    torch.nn.utils.clip_grad_norm_(
                        pet_model.parameters(),
                        max_norm=FITTING_SCHEME.GRADIENT_CLIPPING_MAX_NORM,
                    )
                optim.step()
                optim.zero_grad()

            pet_model.train(False)
            for batch in val_loader:
                if not FITTING_SCHEME.MULTI_GPU:
                    batch.to(device)

                if FITTING_SCHEME.MULTI_GPU:
                    pet_model.module.augmentation = False
                    pet_model.module.create_graph = False
                    predictions_energies, predictions_forces = pet_model(batch)
                else:
                    predictions_energies, predictions_forces = pet_model(
                        batch, augmentation=False, create_graph=False
                    )

                if FITTING_SCHEME.MULTI_GPU:
                    y_list = [el.y for el in batch]
                    batch_y = torch.tensor(
                        y_list, dtype=torch.get_default_dtype(), device=device
                    )

                    n_atoms_list = [el.n_atoms for el in batch]
                    batch_n_atoms = torch.tensor(
                        n_atoms_list, dtype=torch.get_default_dtype(), device=device
                    )
                else:
                    batch_y = batch.y
                    batch_n_atoms = batch.n_atoms

                if FITTING_SCHEME.ENERGIES_LOSS == "per_atom":
                    predictions_energies = predictions_energies / batch_n_atoms
                    ground_truth_energies = batch_y / batch_n_atoms
                else:
                    ground_truth_energies = batch_y

                if MLIP_SETTINGS.USE_ENERGIES:
                    energies_logger.val_logger.update(
                        predictions_energies, ground_truth_energies
                    )
                if MLIP_SETTINGS.USE_FORCES:
                    if FITTING_SCHEME.MULTI_GPU:
                        forces_list = [el.forces for el in batch]
                        batch_forces = torch.cat(forces_list, dim=0).to(device)
                    else:
                        batch_forces = batch.forces
                    forces_logger.val_logger.update(predictions_forces, batch_forces)

            now = {}
            if FITTING_SCHEME.ENERGIES_LOSS == "per_structure":
                energies_key = "energies per structure"
            else:
                energies_key = "energies per atom"

            if MLIP_SETTINGS.USE_ENERGIES:
                now[energies_key] = energies_logger.flush()

            if MLIP_SETTINGS.USE_FORCES:
                now["forces"] = forces_logger.flush()
            now["lr"] = scheduler.get_last_lr()
            now["epoch"] = epoch

            now["elapsed_time"] = time.time() - TIME_TRAINING_STARTED
            now["epoch_time"] = now["elapsed_time"] - last_elapsed_time
            now["estimated_remaining_time"] = (now["elapsed_time"] / epoch) * (
                FITTING_SCHEME.EPOCH_NUM - epoch
            )
            last_elapsed_time = now["elapsed_time"]

            if MLIP_SETTINGS.USE_ENERGIES:
                sliding_energies_rmse = (
                    FITTING_SCHEME.SLIDING_FACTOR * sliding_energies_rmse
                    + (1.0 - FITTING_SCHEME.SLIDING_FACTOR)
                    * now[energies_key]["val"]["rmse"]
                )

                energies_mae_model_keeper.update(
                    pet_model, now[energies_key]["val"]["mae"], epoch
                )
                energies_rmse_model_keeper.update(
                    pet_model, now[energies_key]["val"]["rmse"], epoch
                )

            if MLIP_SETTINGS.USE_FORCES:
                sliding_forces_rmse = (
                    FITTING_SCHEME.SLIDING_FACTOR * sliding_forces_rmse
                    + (1.0 - FITTING_SCHEME.SLIDING_FACTOR)
                    * now["forces"]["val"]["rmse"]
                )
                forces_mae_model_keeper.update(
                    pet_model, now["forces"]["val"]["mae"], epoch
                )
                forces_rmse_model_keeper.update(
                    pet_model, now["forces"]["val"]["rmse"], epoch
                )

            if MLIP_SETTINGS.USE_ENERGIES and MLIP_SETTINGS.USE_FORCES:
                multiplication_mae_model_keeper.update(
                    pet_model,
                    now["forces"]["val"]["mae"] * now[energies_key]["val"]["mae"],
                    epoch,
                    additional_info=[
                        now[energies_key]["val"]["mae"],
                        now["forces"]["val"]["mae"],
                    ],
                )
                multiplication_rmse_model_keeper.update(
                    pet_model,
                    now["forces"]["val"]["rmse"] * now[energies_key]["val"]["rmse"],
                    epoch,
                    additional_info=[
                        now[energies_key]["val"]["rmse"],
                        now["forces"]["val"]["rmse"],
                    ],
                )
            last_lr = scheduler.get_last_lr()[0]
            log_epoch_stats(epoch, FITTING_SCHEME.EPOCH_NUM, now, last_lr, energies_key)

            history.append(now)
            scheduler.step()
            elapsed = time.time() - TIME_SCRIPT_STARTED
            if epoch > 0 and epoch % FITTING_SCHEME.CHECKPOINT_INTERVAL == 0:
                self.epoch = epoch
                pet_checkpoint = {
                    "model_state_dict": pet_model.state_dict(),
                    "optim_state_dict": optim.state_dict(),
                    "scheduler_state_dict": scheduler.state_dict(),
                    "dtype_used": dtype2string(dtype),
                }
                torch.save(
                    pet_checkpoint,
                    f"{checkpoint_dir}/{NAME_OF_CALCULATION}/checkpoint_{epoch}",
                )
                trainer_state_dict = {
                    "optim_state_dict": pet_checkpoint["optim_state_dict"],
                    "scheduler_state_dict": pet_checkpoint["scheduler_state_dict"],
                }
                last_model_state_dict = pet_checkpoint["model_state_dict"]
                if model.is_lora_applied:
                    lora_state_dict = {
                        "lora_rank": FITTING_SCHEME.LORA_RANK,
                        "lora_alpha": FITTING_SCHEME.LORA_ALPHA,
                    }
                else:
                    lora_state_dict = None
                last_model_checkpoint = {
                    "architecture_name": "pet",
                    "trainer_state_dict": trainer_state_dict,
                    "model_state_dict": last_model_state_dict,
                    "best_model_state_dict": self.best_model_state_dict,
                    "best_metric": self.best_metric,
                    "hypers": self.hypers,
                    "epoch": self.epoch,
                    "dataset_info": model.dataset_info,
                    "self_contributions": self_contributions,
                    "lora_state_dict": lora_state_dict,
                }
                torch.save(
                    last_model_checkpoint,
                    f"{checkpoint_dir}/model.ckpt_{epoch}",
                )

            if FITTING_SCHEME.MAX_TIME is not None:
                if elapsed > FITTING_SCHEME.MAX_TIME:
                    logging.info("Reached maximum time\n")
                    break
        logging.info("Training is finished\n")
        logging.info("Saving the model and history...")
        torch.save(
            {
                "model_state_dict": pet_model.state_dict(),
                "optim_state_dict": optim.state_dict(),
                "scheduler_state_dict": scheduler.state_dict(),
                "dtype_used": dtype2string(dtype),
            },
            f"{checkpoint_dir}/{NAME_OF_CALCULATION}/checkpoint",
        )
        with open(f"{checkpoint_dir}/{NAME_OF_CALCULATION}/history.pickle", "wb") as f:
            pickle.dump(history, f)

        def save_model(model_name, model_keeper):
            torch.save(
                model_keeper.best_model.state_dict(),
                f"{checkpoint_dir}/{NAME_OF_CALCULATION}/{model_name}_state_dict",
            )

        summary = ""
        if MLIP_SETTINGS.USE_ENERGIES:
            if FITTING_SCHEME.ENERGIES_LOSS == "per_structure":
                postfix = "per structure"
            if FITTING_SCHEME.ENERGIES_LOSS == "per_atom":
                postfix = "per atom"
            save_model("best_val_mae_energies_model", energies_mae_model_keeper)
            summary += f"best val mae in energies {postfix}: "
            summary += f"{energies_mae_model_keeper.best_error} "
            summary += f"at epoch {energies_mae_model_keeper.best_epoch}\n"

            save_model("best_val_rmse_energies_model", energies_rmse_model_keeper)
            summary += f"best val rmse in energies {postfix}: "
            summary += f"{energies_rmse_model_keeper.best_error} "
            summary += f"at epoch {energies_rmse_model_keeper.best_epoch}\n"

            if energies_mae_model_keeper.best_error < self.best_metric:
                self.best_metric = energies_mae_model_keeper.best_error
                self.best_model_state_dict = (
                    energies_mae_model_keeper.best_model.state_dict()
                )

        if MLIP_SETTINGS.USE_FORCES:
            save_model("best_val_mae_forces_model", forces_mae_model_keeper)
            summary += f"best val mae in forces: {forces_mae_model_keeper.best_error} "
            summary += f"at epoch {forces_mae_model_keeper.best_epoch}\n"

            save_model("best_val_rmse_forces_model", forces_rmse_model_keeper)
            summary += (
                f"best val rmse in forces: {forces_rmse_model_keeper.best_error} "
            )
            summary += f"at epoch {forces_rmse_model_keeper.best_epoch}\n"

            if forces_mae_model_keeper.best_error < self.best_metric:
                self.best_metric = forces_mae_model_keeper.best_error
                self.best_model_state_dict = (
                    forces_mae_model_keeper.best_model.state_dict()
                )

        if MLIP_SETTINGS.USE_ENERGIES and MLIP_SETTINGS.USE_FORCES:
            save_model("best_val_mae_both_model", multiplication_mae_model_keeper)
            summary += f"best both (multiplication) mae in energies {postfix}: "
            summary += (
                f"{multiplication_mae_model_keeper.additional_info[0]} in forces: "
            )
            summary += f"{multiplication_mae_model_keeper.additional_info[1]} "
            summary += f"at epoch {multiplication_mae_model_keeper.best_epoch}\n"

            save_model("best_val_rmse_both_model", multiplication_rmse_model_keeper)
            summary += f"best both (multiplication) rmse in energies {postfix}: "
            summary += (
                f"{multiplication_rmse_model_keeper.additional_info[0]} in forces: "
            )
            summary += (
                f"{multiplication_rmse_model_keeper.additional_info[1]} at epoch "
            )
            summary += f"{multiplication_rmse_model_keeper.best_epoch}\n"

            if multiplication_mae_model_keeper.best_error < self.best_metric:
                self.best_metric = multiplication_mae_model_keeper.best_error
                self.best_model_state_dict = (
                    multiplication_mae_model_keeper.best_model.state_dict()
                )

        with open(f"{checkpoint_dir}/{NAME_OF_CALCULATION}/summary.txt", "wb") as f:
            f.write(summary.encode())
        logging.info(f"Total elapsed time: {time.time() - TIME_SCRIPT_STARTED}")

        ##########################################
        # FINISHING THE PURE PET TRAINING SCRIPT #
        ##########################################
        self.epoch = epoch
        wrapper = load_raw_pet_model(
            self.best_model_state_dict,
            model.hypers,
            all_species,
            self_contributions,
            use_lora_peft=FITTING_SCHEME.USE_LORA_PEFT,
            lora_rank=FITTING_SCHEME.LORA_RANK,
            lora_alpha=FITTING_SCHEME.LORA_ALPHA,
        )
        model.set_trained_model(wrapper)

    def save_checkpoint(self, model, path: Union[str, Path]):
        # This function takes a checkpoint from the PET folder and saves it
        # together with the hypers inside a file that will act as a metatrain
        # checkpoint
        pet_checkpoint = torch.load(
            self.pet_dir / "checkpoint", weights_only=False, map_location="cpu"
        )
        trainer_state_dict = {
            "optim_state_dict": pet_checkpoint["optim_state_dict"],
            "scheduler_state_dict": pet_checkpoint["scheduler_state_dict"],
        }
        last_model_state_dict = pet_checkpoint["model_state_dict"]
        if model.is_lora_applied:
            lora_state_dict = {
                "lora_rank": model.pet.model.rank,
                "lora_alpha": model.pet.model.alpha,
            }
        else:
            lora_state_dict = None
        last_model_checkpoint = {
            "architecture_name": "pet",
            "trainer_state_dict": trainer_state_dict,
            "model_state_dict": last_model_state_dict,
            "best_model_state_dict": self.best_model_state_dict,
            "best_metric": self.best_metric,
            "hypers": self.hypers,
            "epoch": self.epoch,
            "dataset_info": model.dataset_info,
            "self_contributions": model.pet.self_contributions.numpy(),
            "lora_state_dict": lora_state_dict,
        }
        best_model_checkpoint = {
            "architecture_name": "pet",
            "trainer_state_dict": None,
            "model_state_dict": self.best_model_state_dict,
            "best_model_state_dict": None,
            "best_metric": None,
            "hypers": self.hypers,
            "epoch": None,
            "dataset_info": model.dataset_info,
            "self_contributions": model.pet.self_contributions.numpy(),
            "lora_state_dict": lora_state_dict,
        }

        torch.save(
            last_model_checkpoint,
            check_file_extension(f"last_checkpoint_{str(path)}", ".ckpt"),
        )

        torch.save(
            best_model_checkpoint, check_file_extension("best_" + str(path), ".ckpt")
        )

    @classmethod
    def load_checkpoint(cls, path: Union[str, Path], train_hypers) -> "Trainer":
        # This function loads a metatrain PET checkpoint and returns a Trainer
        # instance with the hypers, while also saving the checkpoint in the
        # class
        checkpoint = torch.load(path, weights_only=False, map_location="cpu")
        trainer = cls(train_hypers)
        trainer.pet_trainer_state = checkpoint["trainer_state_dict"]
        trainer.epoch = checkpoint["epoch"]
        old_fitting_scheme = checkpoint["hypers"]["FITTING_SCHEME"]
        new_fitting_scheme = train_hypers
        best_metric = checkpoint["best_metric"]
        best_model_state_dict = checkpoint["best_model_state_dict"]
        # The following code is not reached in the current implementation
        # because the check for the train targets is done earlier in the
        # training process, and changing the training targets between the
        # runs is forbidden. However, this code is kept here for future reference.
        for key in new_fitting_scheme:
            if key in ["USE_ENERGIES", "USE_FORCES"]:
                if new_fitting_scheme[key] != old_fitting_scheme[key]:
                    logging.warning(
                        f"The {key} training hyperparameter was changed from "
                        f"{old_fitting_scheme[key]} to {new_fitting_scheme[key]} "
                        "inbetween the last checkpoint and the current training. "
                        "The `best model` and the `best loss` parts of the checkpoint "
                        "will be reset to avoid inconsistencies."
                    )
                    best_metric = None
                    best_model_state_dict = None
        trainer.best_metric = best_metric
        trainer.best_model_state_dict = best_model_state_dict
        return trainer



import copy
from typing import Optional

import numpy as np
import torch
import torch.nn.functional as F
from torch import nn

from .utilities import NeverRun


class AttentionBlock(nn.Module):
    def __init__(self, total_dim, num_heads, dropout=0.0, epsilon=1e-15):
        super(AttentionBlock, self).__init__()

        self.input_linear = nn.Linear(total_dim, 3 * total_dim)
        self.dropout = nn.Dropout(dropout)
        self.output_linear = nn.Linear(total_dim, total_dim)

        nn.init.xavier_uniform_(self.input_linear.weight)
        nn.init.constant_(self.input_linear.bias, 0.0)
        nn.init.constant_(self.output_linear.bias, 0.0)

        self.num_heads = num_heads
        self.epsilon = epsilon

        if total_dim % num_heads != 0:
            raise ValueError("total dimension is not divisible by the number of heads")
        self.head_dim = total_dim // num_heads
        self.preconditioning = 1.0 / np.sqrt(self.head_dim)

    def forward(self, x, multipliers: Optional[torch.Tensor] = None):
        initial_shape = x.shape
        x = self.input_linear(x)
        x = x.reshape(
            initial_shape[0], initial_shape[1], 3, self.num_heads, self.head_dim
        )
        x = x.permute(2, 0, 3, 1, 4)

        queries, keys, values = x[0], x[1], x[2]
        alpha = torch.matmul(queries, keys.transpose(-2, -1)) * self.preconditioning
        alpha = F.softmax(alpha, dim=-1)
        alpha = self.dropout(alpha)

        if multipliers is not None:
            alpha = alpha * multipliers[:, None, :, :]
            alpha = alpha / (alpha.sum(dim=-1)[..., None] + self.epsilon)

        x = torch.matmul(alpha, values).transpose(1, 2).reshape(initial_shape)
        x = self.output_linear(x)
        return x


class TransformerLayer(torch.nn.Module):
    def __init__(
        self,
        d_model,
        n_heads,
        dim_feedforward=512,
        dropout=0.0,
        activation=F.silu,
        transformer_type="PostLN",
    ):
        super(TransformerLayer, self).__init__()
        self.attention = AttentionBlock(d_model, n_heads, dropout=dropout)

        if transformer_type not in ["PostLN", "PreLN"]:
            raise ValueError("unknown transformer type")
        self.transformer_type = transformer_type
        self.d_model = d_model
        self.norm_attention = nn.LayerNorm(d_model)
        self.norm_mlp = nn.LayerNorm(d_model)
        self.dropout = nn.Dropout(dropout)

        self.activation = activation

        self.mlp = nn.Sequential(
            nn.Linear(d_model, dim_feedforward),
            self.activation,
            nn.Dropout(dropout),
            nn.Linear(dim_feedforward, d_model),
            nn.Dropout(dropout),
        )

    def forward(self, x, multipliers: Optional[torch.Tensor] = None):
        if self.transformer_type == "PostLN":
            x = self.norm_attention(x + self.dropout(self.attention(x, multipliers)))
            x = self.norm_mlp(x + self.mlp(x))
        if self.transformer_type == "PreLN":
            x = x + self.dropout(self.attention(self.norm_attention(x), multipliers))
            x = x + self.mlp(self.norm_mlp(x))
        return x


class Transformer(torch.nn.Module):
    def __init__(self, trans_layer, num_layers):
        super(Transformer, self).__init__()
        self.transformer_type = trans_layer.transformer_type

        self.final_norm = NeverRun()  # for torchscript
        if trans_layer.transformer_type == "PreLN":
            self.final_norm = nn.LayerNorm(trans_layer.d_model)
        self.layers = [copy.deepcopy(trans_layer) for _ in range(num_layers)]
        self.layers = nn.ModuleList(self.layers)

    def forward(self, x: torch.Tensor, multipliers: Optional[torch.Tensor] = None):
        for layer in self.layers:
            x = layer(x, multipliers)
        if self.transformer_type == "PreLN":
            x = self.final_norm(x)
        return x



import logging
import warnings
from pathlib import PurePath
from typing import IO, List, Tuple, Union

import ase.io
import torch
from ase.stress import voigt_6_to_full_3x3_stress
from metatensor.torch import Labels, TensorBlock, TensorMap
from metatensor.torch.atomistic import System, systems_to_torch
from omegaconf import DictConfig

from ..target_info import TargetInfo, get_energy_target_info, get_generic_target_info


def read(filename: Union[str, PurePath, IO], *args, **kwargs) -> List[ase.Atoms]:
    """Wrapper around the :func:`ase.io.read` function.

    The wrapper provides a more informative error message in case of failure.
    Additionally, it will make the keys ``"energy"``, ``"forces"`` and ``"stress"``
    available from the calculator and the info/arrays dictionary.

    .. warning::

        Lists of atoms read with this function can NOT be written back to a file with
        :func:`ase.io.write` because of the duplicated keys.

    :param filename: Name of the file to read from or a file descriptor.
    :param args: additional positional arguments for :func:`ase.io.read`
    :param kwargs: additional keyword arguments for :func:`ase.io.read`
    :returns: A list of atoms
    """
    try:
        frames = ase.io.read(filename, *args, **kwargs)
    except Exception as e:
        raise ValueError(f"Failed to read '{filename}' with ASE: {e}") from e

    # allow access of "special" keys from calculator and `info`/`arrays` dictionary
    for atoms in frames:
        if hasattr(atoms, "calc") and atoms.calc is not None:
            results = atoms.calc.results
            if "energy" in results:
                atoms.info["energy"] = results["energy"]
            if "forces" in results:
                atoms.arrays["forces"] = results["forces"]
            if "stress" in results:
                atoms.info["stress"] = voigt_6_to_full_3x3_stress(results["stress"])

    return frames


def read_systems(filename: str) -> List[System]:
    """Store system informations using ase.

    :param filename: name of the file to read
    :returns: A list of systems
    """
    return systems_to_torch(read(filename, ":"), dtype=torch.float64)


def _read_energy_ase(filename: str, key: str) -> List[TensorBlock]:
    """Store energy information in a List of :class:`metatensor.TensorBlock`.

    :param filename: name of the file to read
    :param key: target value key name to be parsed from the file.
    :returns: TensorMap containing the energies
    """
    frames = read(filename, ":")

    properties = Labels("energy", torch.tensor([[0]]))

    blocks = []
    for i_system, atoms in enumerate(frames):
        if key not in atoms.info:
            raise ValueError(
                f"energy key {key!r} was not found in system {filename!r} at index "
                f"{i_system}"
            )

        values = torch.tensor([[atoms.info[key]]], dtype=torch.float64)
        samples = Labels(["system"], torch.tensor([[i_system]]))

        block = TensorBlock(
            values=values,
            samples=samples,
            components=[],
            properties=properties,
        )
        blocks.append(block)

    return blocks


def _read_forces_ase(filename: str, key: str = "forces") -> List[TensorBlock]:
    """Store force information in a List of :class:`metatensor.TensorBlock` which can be
    used as ``position`` gradients.

    :param filename: name of the file to read
    :param key: target value key name to be parsed from the file.
    :returns: TensorMap containing the forces
    """
    frames = read(filename, ":")

    components = [Labels(["xyz"], torch.arange(3).reshape(-1, 1))]
    properties = Labels("energy", torch.tensor([[0]]))

    blocks = []
    for i_system, atoms in enumerate(frames):
        if key not in atoms.arrays:
            raise ValueError(
                f"forces key {key!r} was not found in system {filename!r} at index "
                f"{i_system}"
            )

        # We store forces as positions gradients which means we invert the sign
        values = -torch.tensor(atoms.arrays[key], dtype=torch.float64)
        values = values.reshape(-1, 3, 1)

        samples = Labels(
            ["sample", "system", "atom"],
            torch.tensor([[0, i_system, a] for a in range(len(values))]),
        )

        block = TensorBlock(
            values=values,
            samples=samples,
            components=components,
            properties=properties,
        )

        blocks.append(block)

    return blocks


def _read_virial_ase(filename: str, key: str = "virial") -> List[TensorBlock]:
    """Store virial information in a List of :class:`metatensor.TensorBlock` which can
    be used as ``strain`` gradients.

    :param filename: name of the file to read
    :param key: target value key name to be parsed from the file
    :returns: TensorMap containing the virial
    """
    return _read_virial_stress_ase(filename=filename, key=key, is_virial=True)


def _read_stress_ase(filename: str, key: str = "stress") -> List[TensorBlock]:
    """Store stress information in a List of :class:`metatensor.TensorBlock` which can
    be used as ``strain`` gradients.

    :param filename: name of the file to read
    :param key: target value key name to be parsed from the file
    :returns: TensorMap containing the stress
    """
    return _read_virial_stress_ase(filename=filename, key=key, is_virial=False)


def _read_virial_stress_ase(
    filename: str, key: str, is_virial: bool = True
) -> List[TensorBlock]:
    frames = read(filename, ":")

    samples = Labels(["sample"], torch.tensor([[0]]))
    components = [
        Labels(["xyz_1"], torch.arange(3).reshape(-1, 1)),
        Labels(["xyz_2"], torch.arange(3).reshape(-1, 1)),
    ]
    properties = Labels("energy", torch.tensor([[0]]))

    blocks = []
    for i_system, atoms in enumerate(frames):
        if key not in atoms.info:
            target_name = "virial" if is_virial else "stress"
            raise ValueError(
                f"{target_name} key {key!r} was not found in system {filename!r} at "
                f"index {i_system}"
            )

        values = torch.tensor(atoms.info[key].tolist(), dtype=torch.float64)

        if values.shape == (9,):
            warnings.warn(
                "Found 9-long numerical vector for the stress/virial in system "
                f"{i_system}. Assume a row major format for the conversion into a "
                "3 x 3 matrix.",
                stacklevel=2,
            )
        elif values.shape != (3, 3):
            raise ValueError(
                f"Values in system {i_system} has shape {values.shape}. "
                "Stress/virial must be a 3 x 3 matrix or a 9-long numerical vector."
            )

        values = values.reshape(-1, 3, 3, 1)

        if is_virial:
            values *= -1
        else:  # is stress
            if atoms.cell.volume == 0:
                raise ValueError(
                    f"system {i_system} has zero cell vectors. Stress can only "
                    "be used if cell is non zero."
                )
            values *= atoms.cell.volume

        block = TensorBlock(
            values=values,
            samples=samples,
            components=components,
            properties=properties,
        )
        blocks.append(block)

    return blocks


def read_energy(target: DictConfig) -> Tuple[List[TensorMap], TargetInfo]:
    target_key = target["key"]

    blocks = _read_energy_ase(
        filename=target["read_from"],
        key=target["key"],
    )

    add_position_gradients = False
    if target["forces"]:
        try:
            position_gradients = _read_forces_ase(
                filename=target["forces"]["read_from"],
                key=target["forces"]["key"],
            )
        except Exception:
            logging.warning(f"No forces found in section {target_key!r}.")
            add_position_gradients = False
        else:
            logging.info(
                f"Forces found in section {target_key!r}, "
                "we will use this gradient to train the model"
            )
            for block, position_gradient in zip(blocks, position_gradients):
                block.add_gradient(parameter="positions", gradient=position_gradient)
            add_position_gradients = True

    if target["stress"] and target["virial"]:
        raise ValueError("Cannot use stress and virial at the same time")

    add_strain_gradients = False

    if target["stress"]:
        try:
            strain_gradients = _read_stress_ase(
                filename=target["stress"]["read_from"],
                key=target["stress"]["key"],
            )
        except Exception:
            logging.warning(f"No stress found in section {target_key!r}.")
            add_strain_gradients = False
        else:
            logging.info(
                f"Stress found in section {target_key!r}, "
                "we will use this gradient to train the model"
            )
            for block, strain_gradient in zip(blocks, strain_gradients):
                block.add_gradient(parameter="strain", gradient=strain_gradient)
            add_strain_gradients = True

    if target["virial"]:
        try:
            strain_gradients = _read_virial_ase(
                filename=target["virial"]["read_from"],
                key=target["virial"]["key"],
            )
        except Exception:
            logging.warning(f"No virial found in section {target_key!r}.")
            add_strain_gradients = False
        else:
            logging.info(
                f"Virial found in section {target_key!r}, "
                "we will use this gradient to train the model"
            )
            for block, strain_gradient in zip(blocks, strain_gradients):
                block.add_gradient(parameter="strain", gradient=strain_gradient)
            add_strain_gradients = True
    tensor_map_list = [
        TensorMap(
            keys=Labels(["_"], torch.tensor([[0]])),
            blocks=[block],
        )
        for block in blocks
    ]
    target_info = get_energy_target_info(
        target, add_position_gradients, add_strain_gradients
    )
    return tensor_map_list, target_info


def read_generic(target: DictConfig) -> Tuple[List[TensorMap], TargetInfo]:
    filename = target["read_from"]
    frames = read(filename, ":")

    # we don't allow ASE to read spherical tensors with more than one irrep,
    # otherwise it's a mess
    if (
        isinstance(target["type"], DictConfig)
        and next(iter(target["type"].keys())) == "spherical"
    ):
        irreps = target["type"]["spherical"]["irreps"]
        if len(irreps) > 1:
            raise ValueError(
                "The metatrain ASE reader does not support reading "
                "spherical tensors with more than one irreducible "
                "representation. Please use the metatensor reader."
            )

    target_info = get_generic_target_info(target)
    components = target_info.layout.block().components
    properties = target_info.layout.block().properties
    shape_after_samples = target_info.layout.block().shape[1:]
    per_atom = target_info.per_atom
    keys = target_info.layout.keys

    target_key = target["key"]

    tensor_maps = []
    for i_system, atoms in enumerate(frames):
        if (per_atom and target_key not in atoms.arrays) or (
            not per_atom and target_key not in atoms.info
        ):
            raise ValueError(
                f"Target key {target_key!r} was not found in system {filename!r} at "
                f"index {i_system}"
            )

        if per_atom:
            data = atoms.arrays[target_key]
        else:
            data = atoms.info[target_key]

        # here we reshape to allow for more flexibility; this is actually
        # necessary for the `arrays`, which are stored in a 2D array
        values = torch.tensor(data, dtype=torch.float64).reshape(
            [-1] + shape_after_samples
        )

        samples = (
            Labels(
                ["system", "atom"],
                torch.tensor([[i_system, a] for a in range(len(values))]),
            )
            if per_atom
            else Labels(
                ["system"],
                torch.tensor([[i_system]]),
            )
        )

        block = TensorBlock(
            values=values,
            samples=samples,
            components=components,
            properties=properties,
        )
        tensor_map = TensorMap(
            keys=keys,
            blocks=[block],
        )
        tensor_maps.append(tensor_map)

    return tensor_maps, target_info



import random

import numpy as np
import torch
from metatensor.torch.atomistic import ModelOutput
from omegaconf import OmegaConf

from metatrain.soap_bpnn import SoapBpnn, Trainer
from metatrain.utils.data import Dataset, DatasetInfo
from metatrain.utils.data.readers import read_systems, read_targets
from metatrain.utils.data.target_info import get_energy_target_info

from . import DATASET_PATH, DEFAULT_HYPERS, MODEL_HYPERS


# reproducibility
random.seed(0)
np.random.seed(0)
torch.manual_seed(0)


def test_regression_init():
    """Perform a regression test on the model at initialization"""

    targets = {}
    targets["mtt::U0"] = get_energy_target_info({"unit": "eV"})

    dataset_info = DatasetInfo(
        length_unit="Angstrom", atomic_types=[1, 6, 7, 8], targets=targets
    )
    model = SoapBpnn(MODEL_HYPERS, dataset_info)

    # Predict on the first five systems
    systems = read_systems(DATASET_PATH)[:5]
    systems = [system.to(torch.float32) for system in systems]

    output = model(
        systems,
        {"mtt::U0": ModelOutput(quantity="energy", unit="", per_atom=False)},
    )

    expected_output = torch.tensor(
        [
            [-0.067431554198],
            [-0.008789218962],
            [-0.108400978148],
            [-0.067436940968],
            [-0.021975118667],
        ]
    )

    # if you need to change the hardcoded values:
    # torch.set_printoptions(precision=12)
    # print(output["mtt::U0"].block().values)

    torch.testing.assert_close(output["mtt::U0"].block().values, expected_output)


def test_regression_train():
    """Perform a regression test on the model when
    trained for 2 epoch on a small dataset"""

    systems = read_systems(DATASET_PATH)

    conf = {
        "mtt::U0": {
            "quantity": "energy",
            "read_from": DATASET_PATH,
            "reader": "ase",
            "key": "U0",
            "unit": "eV",
            "type": "scalar",
            "per_atom": False,
            "num_subtargets": 1,
            "forces": False,
            "stress": False,
            "virial": False,
        }
    }
    targets, target_info_dict = read_targets(OmegaConf.create(conf))
    dataset = Dataset.from_dict({"system": systems, "mtt::U0": targets["mtt::U0"]})

    hypers = DEFAULT_HYPERS.copy()
    hypers["training"]["num_epochs"] = 2

    dataset_info = DatasetInfo(
        length_unit="Angstrom", atomic_types=[1, 6, 7, 8], targets=target_info_dict
    )
    model = SoapBpnn(MODEL_HYPERS, dataset_info)

    hypers["training"]["num_epochs"] = 1
    trainer = Trainer(hypers["training"])
    trainer.train(
        model=model,
        dtype=torch.float32,
        devices=[torch.device("cpu")],
        train_datasets=[dataset],
        val_datasets=[dataset],
        checkpoint_dir=".",
    )

    # Predict on the first five systems
    systems = [system.to(torch.float32) for system in systems]
    output = model(
        systems[:5],
        {"mtt::U0": ModelOutput(quantity="energy", unit="", per_atom=False)},
    )

    expected_output = torch.tensor(
        [
            [0.751162171364],
            [0.127142444253],
            [1.123749136925],
            [1.653730630875],
            [1.415526270866],
        ]
    )

    # if you need to change the hardcoded values:
    # torch.set_printoptions(precision=12)
    # print(output["mtt::U0"].block().values)

    torch.testing.assert_close(output["mtt::U0"].block().values, expected_output)



from .combine_dataloaders import CombinedDataLoader  # noqa: F401
from .dataset import (  # noqa: F401
    Dataset,
    DatasetInfo,
    DiskDataset,
    DiskDatasetWriter,
    _is_disk_dataset,
    check_datasets,
    collate_fn,
    get_all_targets,
    get_atomic_types,
    get_stats,
)
from .get_dataset import get_dataset  # noqa: F401
from .readers import read_systems, read_targets  # noqa: F401
from .system_to_ase import system_to_ase  # noqa: F401
from .target_info import TargetInfo  # noqa: F401
from .writers import write_predictions  # noqa: F401



import importlib
import warnings
from pathlib import Path
from typing import Dict, List, Optional, Tuple

from metatensor.torch import TensorMap
from metatensor.torch.atomistic import System
from omegaconf import DictConfig

from ..target_info import TargetInfo


AVAILABLE_READERS = ["ase", "metatensor"]
""":py:class:`list`: list containing all implemented reader libraries"""

DEFAULT_READER = {".xyz": "ase", ".extxyz": "ase", ".mts": "metatensor"}
""":py:class:`dict`: dictionary mapping file extensions to a default reader"""


def read_systems(
    filename: str,
    reader: Optional[str] = None,
) -> List[System]:
    """Read system informations from a file.

    :param filename: name of the file to read
    :param reader: reader library for parsing the file. If :py:obj:`None` the library is
        is tried to determined from the file extension.
    :param dtype: desired data type of returned tensor
    :returns: list of systems
        determined from the file extension.
    :returns: list of systems stored in double precision
    """
    if reader is None:
        try:
            file_suffix = Path(filename).suffix
            reader = DEFAULT_READER[file_suffix]
        except KeyError:
            raise ValueError(
                f"File extension {file_suffix!r} is not linked to a default reader "
                "library. You can try reading it by setting a specific 'reader' from "
                f"the known ones: {', '.join(AVAILABLE_READERS)} "
            )

    try:
        reader_mod = importlib.import_module(
            name=f".{reader}", package="metatrain.utils.data.readers"
        )
    except ImportError:
        raise ValueError(
            f"Reader library {reader!r} not supported. Choose from "
            f"{', '.join(AVAILABLE_READERS)}"
        )

    try:
        reader_met = reader_mod.read_systems
    except AttributeError:
        raise ValueError(
            f"Reader library {reader!r} cannot read systems."
            f"You can try with other readers: {AVAILABLE_READERS}"
        )

    systems = reader_met(filename)

    # elements in data are `torch.ScriptObject`s and their `dtype` is an integer.
    # A C++ double/torch.float64 is `7` according to
    # https://github.com/pytorch/pytorch/blob/207564bab1c4fe42750931765734ee604032fb69/c10/core/ScalarType.h#L54-L93
    if not all(s.dtype == 7 for s in systems):
        raise ValueError("The loaded systems are not in double precision.")

    return systems


def read_targets(
    conf: DictConfig,
) -> Tuple[Dict[str, List[TensorMap]], Dict[str, TargetInfo]]:
    """Reading all target information from a fully expanded config.

    To get such a config you can use :func:`expand_dataset_config
    <metatrain.utils.omegaconf.expand_dataset_config>`. All targets are stored in double
    precision.

    This function uses subfunctions like :func:`read_energy` to parse the requested
    target quantity. Currently only `energy` is a supported target property. But, within
    the `energy` section gradients such as `forces`, the `stress` or the `virial` can be
    added. Other gradients are silently ignored.

    :param conf: config containing the keys for what should be read.
    :returns: Dictionary containing a list of TensorMaps for each target section in the
        config as well as a ``Dict[str, TargetInfo]`` object
        containing the metadata of the targets.

    :raises ValueError: if the target name is not valid. Valid target names are
        those that either start with ``mtt::`` or those that are in the list of
        standard outputs of ``metatensor.torch.atomistic`` (see
        https://docs.metatensor.org/latest/atomistic/outputs.html)
    """
    target_dictionary = {}
    target_info_dictionary = {}
    standard_outputs_list = [
        "energy",
        "non_conservative_forces",
        "non_conservative_stress",
    ]

    for target_key, target in conf.items():
        is_standard_target = target_key in standard_outputs_list
        if not is_standard_target and not target_key.startswith("mtt::"):
            if target_key.lower() in ["force", "forces", "virial", "stress"]:
                warnings.warn(
                    f"{target_key!r} should not be its own top-level target, "
                    "but rather a sub-section of the 'energy' target",
                    stacklevel=2,
                )
            else:
                raise ValueError(
                    f"Target name ({target_key}) must either be one of "
                    f"{standard_outputs_list} or start with `mtt::`."
                )
        if (
            "force" in target_key.lower()
            or "virial" in target_key.lower()
            or "stress" in target_key.lower()
        ):
            warnings.warn(
                f"the name of {target_key!r} resembles to a gradient of "
                "energies; it should probably not be its own top-level target, "
                "but rather a gradient sub-section of a target with the "
                "`energy` quantity",
                stacklevel=2,
            )

        is_energy = (
            (target["quantity"] == "energy")
            and (not target["per_atom"])
            and target["num_subtargets"] == 1
            and target["type"] == "scalar"
        )
        energy_or_generic = "energy" if is_energy else "generic"

        reader = target["reader"]
        filename = target["read_from"]

        if reader is None:
            try:
                file_suffix = Path(filename).suffix
                reader = DEFAULT_READER[file_suffix]
            except KeyError:
                raise ValueError(
                    f"File extension {file_suffix!r} is not linked to a default reader "
                    "library. You can try reading it by setting a specific 'reader' "
                    f"from the known ones: {', '.join(AVAILABLE_READERS)} "
                )

        try:
            reader_mod = importlib.import_module(
                name=f".{reader}", package="metatrain.utils.data.readers"
            )
        except ImportError:
            raise ValueError(
                f"Reader library {reader!r} not supported. Choose from "
                f"{', '.join(AVAILABLE_READERS)}"
            )

        try:
            reader_met = getattr(reader_mod, f"read_{energy_or_generic}")
        except AttributeError:
            raise ValueError(
                f"Reader library {reader!r} cannot read {target!r}."
                f"You can try with other readers: {AVAILABLE_READERS}"
            )

        targets_as_list_of_tensor_maps, target_info = reader_met(target)

        # elements in data are `torch.ScriptObject`s and their `dtype` is an integer.
        # A C++ double/torch.float64 is `7` according to
        # https://github.com/pytorch/pytorch/blob/207564bab1c4fe42750931765734ee604032fb69/c10/core/ScalarType.h#L54-L93
        if not all(t.dtype == 7 for t in targets_as_list_of_tensor_maps):
            raise ValueError("The loaded targets are not in double precision.")

        target_dictionary[target_key] = targets_as_list_of_tensor_maps
        target_info_dictionary[target_key] = target_info

    return target_dictionary, target_info_dictionary



from typing import List, Union

import metatensor.torch
import torch
from metatensor.torch import Labels, TensorBlock, TensorMap
from omegaconf import DictConfig


class TargetInfo:
    """A class that contains information about a target.

    :param quantity: The physical quantity of the target (e.g., "energy").
    :param layout: The layout of the target, as a ``TensorMap`` with 0 samples.
        This ``TensorMap`` will be used to retrieve the names of
        the ``samples``, as well as the ``components`` and ``properties`` of the
        target and their gradients. For example, this allows to infer the type of
        the target (scalar, Cartesian tensor, spherical tensor), whether it is per
        atom, the names of its gradients, etc.
    :param unit: The unit of the target. If :py:obj:`None` the ``unit`` will be set to
        an empty string ``""``.
    """

    def __init__(
        self,
        quantity: str,
        layout: TensorMap,
        unit: Union[None, str] = "",
    ):
        # one of these will be set to True inside the _check_layout method
        self.is_scalar = False
        self.is_cartesian = False
        self.is_spherical = False

        self._check_layout(layout)

        self.quantity = quantity  # float64: otherwise metatensor can't serialize
        self.layout = layout
        self.unit = unit if unit is not None else ""

    @property
    def gradients(self) -> List[str]:
        """Sorted and unique list of gradient names."""
        if self.is_scalar:
            return sorted(self.layout.block().gradients_list())
        else:
            return []

    @property
    def per_atom(self) -> bool:
        """Whether the target is per atom."""
        return "atom" in self.layout.block(0).samples.names

    def __repr__(self):
        return (
            f"TargetInfo(quantity={self.quantity!r}, unit={self.unit!r}, "
            f"layout={self.layout!r})"
        )

    def __eq__(self, other):
        if not isinstance(other, TargetInfo):
            raise NotImplementedError(
                "Comparison between a TargetInfo instance and a "
                f"{type(other).__name__} instance is not implemented."
            )
        return (
            self.quantity == other.quantity
            and self.unit == other.unit
            and metatensor.torch.equal(self.layout, other.layout)
        )

    def _check_layout(self, layout: TensorMap) -> None:
        """Check that the layout is a valid layout."""

        # examine basic properties of all blocks
        for block in layout.blocks():
            for sample_name in block.samples.names:
                if sample_name not in ["system", "atom"]:
                    raise ValueError(
                        "The layout ``TensorMap`` of a target should only have samples "
                        "named 'system' or 'atom', but found "
                        f"'{sample_name}' instead."
                    )
            if len(block.values) != 0:
                raise ValueError(
                    "The layout ``TensorMap`` of a target should have 0 "
                    f"samples, but found {len(block.values)} samples."
                )

        # examine the components of the first block to decide whether this is
        # a scalar, a Cartesian tensor or a spherical tensor

        if len(layout) == 0:
            raise ValueError(
                "The layout ``TensorMap`` of a target should have at least one "
                "block, but found 0 blocks."
            )
        components_first_block = layout.block(0).components
        if len(components_first_block) == 0:
            self.is_scalar = True
        elif components_first_block[0].names[0].startswith("xyz"):
            self.is_cartesian = True
        elif (
            len(components_first_block) == 1
            and components_first_block[0].names[0] == "o3_mu"
        ):
            self.is_spherical = True
        else:
            raise ValueError(
                "The layout ``TensorMap`` of a target should be "
                "either scalars, Cartesian tensors or spherical tensors. The type of "
                "the target could not be determined."
            )

        if self.is_scalar:
            if layout.keys.names != ["_"]:
                raise ValueError(
                    "The layout ``TensorMap`` of a scalar target should have "
                    "a single key sample named '_'."
                )
            if len(layout.blocks()) != 1:
                raise ValueError(
                    "The layout ``TensorMap`` of a scalar target should have "
                    "a single block."
                )
            gradients_names = layout.block(0).gradients_list()
            for gradient_name in gradients_names:
                if gradient_name not in ["positions", "strain"]:
                    raise ValueError(
                        "Only `positions` and `strain` gradients are supported for "
                        "scalar targets. "
                        f"Found '{gradient_name}' instead."
                    )
        if self.is_cartesian:
            if layout.keys.names != ["_"]:
                raise ValueError(
                    "The layout ``TensorMap`` of a Cartesian tensor target should have "
                    "a single key sample named '_'."
                )
            if len(layout.blocks()) != 1:
                raise ValueError(
                    "The layout ``TensorMap`` of a Cartesian tensor target should have "
                    "a single block."
                )
            if len(layout.block(0).gradients_list()) > 0:
                raise ValueError(
                    "Gradients of Cartesian tensor targets are not supported."
                )

        if self.is_spherical:
            if layout.keys.names != ["o3_lambda", "o3_sigma"]:
                raise ValueError(
                    "The layout ``TensorMap`` of a spherical tensor target "
                    "should have  two keys named 'o3_lambda' and 'o3_sigma'."
                    f"Found '{layout.keys.names}' instead."
                )
            for key, block in layout.items():
                o3_lambda, o3_sigma = (
                    int(key.values[0].item()),
                    int(key.values[1].item()),
                )
                if o3_sigma not in [-1, 1]:
                    raise ValueError(
                        "The layout ``TensorMap`` of a spherical tensor target should "
                        "have a key sample 'o3_sigma' that is either -1 or 1."
                        f"Found '{o3_sigma}' instead."
                    )
                if o3_lambda < 0:
                    raise ValueError(
                        "The layout ``TensorMap`` of a spherical tensor target should "
                        "have a key sample 'o3_lambda' that is non-negative."
                        f"Found '{o3_lambda}' instead."
                    )
                components = block.components
                if len(components) != 1:
                    raise ValueError(
                        "The layout ``TensorMap`` of a spherical tensor target should "
                        "have a single component."
                    )
                if len(components[0]) != 2 * o3_lambda + 1:
                    raise ValueError(
                        "Each ``TensorBlock`` of a spherical tensor target should have "
                        "a component with 2*o3_lambda + 1 elements."
                        f"Found '{len(components[0])}' elements instead."
                    )
                if len(block.gradients_list()) > 0:
                    raise ValueError(
                        "Gradients of spherical tensor targets are not supported."
                    )

    def is_compatible_with(self, other: "TargetInfo") -> bool:
        """Check if two targets are compatible.

        Two target infos are compatible if they have the same quantity, unit,
        and layout, except for gradients. This method can be used to check if two
        target infos with the same name can correspond to the same output
        in a model.

        :param other: The target info to compare with.
        :return: :py:obj:`True` if the two target infos are compatible,
            :py:obj:`False` otherwise.
        """
        if self.quantity != other.quantity:
            return False
        if self.unit != other.unit:
            return False
        if self.layout.keys.names != other.layout.keys.names:
            return False
        for key, block in self.layout.items():
            if key not in other.layout.keys:
                return False
            other_block = other.layout[key]
            if not block.samples == other_block.samples:
                return False
            if not block.components == other_block.components:
                return False
            if not block.properties == other_block.properties:
                return False
            # gradients are not checked on purpose
        return True


def get_energy_target_info(
    target: DictConfig,
    add_position_gradients: bool = False,
    add_strain_gradients: bool = False,
) -> TargetInfo:
    block = TensorBlock(
        # float64: otherwise metatensor can't serialize
        values=torch.empty(0, 1, dtype=torch.float64),
        samples=Labels(
            names=["system"],
            values=torch.empty((0, 1), dtype=torch.int32),
        ),
        components=[],
        properties=Labels.range("energy", 1),
    )

    if add_position_gradients:
        position_gradient_block = TensorBlock(
            # float64: otherwise metatensor can't serialize
            values=torch.empty(0, 3, 1, dtype=torch.float64),
            samples=Labels(
                names=["sample", "atom"],
                values=torch.empty((0, 2), dtype=torch.int32),
            ),
            components=[
                Labels(
                    names=["xyz"],
                    values=torch.arange(3, dtype=torch.int32).reshape(-1, 1),
                ),
            ],
            properties=Labels.range("energy", 1),
        )
        block.add_gradient("positions", position_gradient_block)

    if add_strain_gradients:
        strain_gradient_block = TensorBlock(
            # float64: otherwise metatensor can't serialize
            values=torch.empty(0, 3, 3, 1, dtype=torch.float64),
            samples=Labels(
                names=["sample", "atom"],
                values=torch.empty((0, 2), dtype=torch.int32),
            ),
            components=[
                Labels(
                    names=["xyz_1"],
                    values=torch.arange(3, dtype=torch.int32).reshape(-1, 1),
                ),
                Labels(
                    names=["xyz_2"],
                    values=torch.arange(3, dtype=torch.int32).reshape(-1, 1),
                ),
            ],
            properties=Labels.range("energy", 1),
        )
        block.add_gradient("strain", strain_gradient_block)

    layout = TensorMap(
        keys=Labels.single(),
        blocks=[block],
    )

    target_info = TargetInfo(
        quantity="energy",
        unit=target["unit"],
        layout=layout,
    )
    return target_info


def get_generic_target_info(target: DictConfig) -> TargetInfo:
    if target["type"] == "scalar":
        return _get_scalar_target_info(target)
    elif len(target["type"]) == 1 and next(iter(target["type"])).lower() == "cartesian":
        return _get_cartesian_target_info(target)
    elif len(target["type"]) == 1 and next(iter(target["type"])) == "spherical":
        return _get_spherical_target_info(target)
    else:
        raise ValueError(
            f"Target type {target['type']} is not supported. "
            "Supported types are 'scalar', 'cartesian' and 'spherical'."
        )


def _get_scalar_target_info(target: DictConfig) -> TargetInfo:
    sample_names = ["system"]
    if target["per_atom"]:
        sample_names.append("atom")

    block = TensorBlock(
        # float64: otherwise metatensor can't serialize
        values=torch.empty(0, target["num_subtargets"], dtype=torch.float64),
        samples=Labels(
            names=sample_names,
            values=torch.empty((0, len(sample_names)), dtype=torch.int32),
        ),
        components=[],
        properties=Labels.range("properties", target["num_subtargets"]),
    )
    layout = TensorMap(
        keys=Labels.single(),
        blocks=[block],
    )

    target_info = TargetInfo(
        quantity=target["quantity"],
        unit=target["unit"],
        layout=layout,
    )
    return target_info


def _get_cartesian_target_info(target: DictConfig) -> TargetInfo:
    sample_names = ["system"]
    if target["per_atom"]:
        sample_names.append("atom")

    cartesian_key = next(iter(target["type"]))

    if target["type"][cartesian_key]["rank"] == 1:
        components = [Labels(["xyz"], torch.arange(3).reshape(-1, 1))]
    else:
        components = []
        for component in range(1, target["type"][cartesian_key]["rank"] + 1):
            components.append(
                Labels(
                    names=[f"xyz_{component}"],
                    values=torch.arange(3, dtype=torch.int32).reshape(-1, 1),
                )
            )

    block = TensorBlock(
        # float64: otherwise metatensor can't serialize
        values=torch.empty(
            [0] + [3] * len(components) + [target["num_subtargets"]],
            dtype=torch.float64,
        ),
        samples=Labels(
            names=sample_names,
            values=torch.empty((0, len(sample_names)), dtype=torch.int32),
        ),
        components=components,
        properties=Labels.range("properties", target["num_subtargets"]),
    )
    layout = TensorMap(
        keys=Labels.single(),
        blocks=[block],
    )

    target_info = TargetInfo(
        quantity=target["quantity"],
        unit=target["unit"],
        layout=layout,
    )
    return target_info


def _get_spherical_target_info(target: DictConfig) -> TargetInfo:
    sample_names = ["system"]
    if target["per_atom"]:
        sample_names.append("atom")

    irreps = target["type"]["spherical"]["irreps"]
    keys = []
    blocks = []
    for irrep in irreps:
        components = [
            Labels(
                names=["o3_mu"],
                values=torch.arange(
                    -irrep["o3_lambda"], irrep["o3_lambda"] + 1, dtype=torch.int32
                ).reshape(-1, 1),
            )
        ]
        block = TensorBlock(
            # float64: otherwise metatensor can't serialize
            values=torch.empty(
                0,
                2 * irrep["o3_lambda"] + 1,
                target["num_subtargets"],
                dtype=torch.float64,
            ),
            samples=Labels(
                names=sample_names,
                values=torch.empty((0, len(sample_names)), dtype=torch.int32),
            ),
            components=components,
            properties=Labels.range("properties", target["num_subtargets"]),
        )
        keys.append([irrep["o3_lambda"], irrep["o3_sigma"]])
        blocks.append(block)

    layout = TensorMap(
        keys=Labels(["o3_lambda", "o3_sigma"], torch.tensor(keys, dtype=torch.int32)),
        blocks=blocks,
    )

    target_info = TargetInfo(
        quantity=target["quantity"],
        unit=target["unit"],
        layout=layout,
    )
    return target_info


def is_auxiliary_output(name: str) -> bool:
    is_auxiliary = (
        name == "features" or name == "energy_ensemble" or name.startswith("mtt::aux::")
    )
    return is_auxiliary
